from typing import Any

from eventsourcing.domain.model.array import (
    AbstractArrayRepository,
    AbstractBigArrayRepository,
)
from eventsourcing.infrastructure.eventsourcedrepository import EventSourcedRepository


class ArrayRepository(AbstractArrayRepository, EventSourcedRepository):
    pass


class BigArrayRepository(AbstractBigArrayRepository, EventSourcedRepository):
    subrepo_class = ArrayRepository

    def __init__(self, array_size: int = 10000, *args: Any, **kwargs: Any):
        super(BigArrayRepository, self).__init__(*args, **kwargs)
        self._subrepo = self.subrepo_class(
            event_store=self.event_store, array_size=array_size
        )

    @property
    def subrepo(self) -> ArrayRepository:
        return self._subrepo



from eventsourcing.domain.model.decorators import attribute
from eventsourcing.domain.model.entity import (
    EntityWithHashchain,
    TDomainEntity,
    TimestampedVersionedEntity,
)
from eventsourcing.domain.model.repository import AbstractEntityRepository


class Example(EntityWithHashchain, TimestampedVersionedEntity):
    """
    An example event sourced domain model entity.
    """

    class Event(
        EntityWithHashchain.Event[TDomainEntity],
        TimestampedVersionedEntity.Event[TDomainEntity],
    ):
        """Supertype for events of example entities."""

    class Created(
        Event,
        EntityWithHashchain.Created[TDomainEntity],
        TimestampedVersionedEntity.Created[TDomainEntity],
    ):
        """Published when an Example is created."""

    class AttributeChanged(
        Event[TDomainEntity], TimestampedVersionedEntity.AttributeChanged[TDomainEntity]
    ):
        """Published when an Example is created."""

    class Discarded(
        Event[TDomainEntity], TimestampedVersionedEntity.Discarded[TDomainEntity]
    ):
        """Published when an Example is discarded."""

    class Heartbeat(
        Event[TDomainEntity], TimestampedVersionedEntity.Event[TDomainEntity]
    ):
        """Published when a heartbeat in the entity occurs (see below)."""

        def mutate(self, obj: "Example") -> None:
            """Updates 'obj' with values from self."""
            obj._count_heartbeats += 1

    def __init__(self, foo="", a="", b="", **kwargs):
        super(Example, self).__init__(**kwargs)
        self._foo = foo
        self._a = a
        self._b = b
        self._count_heartbeats = 0

    @attribute
    def foo(self):
        """An example attribute."""

    @attribute
    def a(self):
        """An example attribute."""

    @attribute
    def b(self):
        """Another example attribute."""

    def beat_heart(self, number_of_beats=1):
        self.__assert_not_discarded__()
        while number_of_beats > 0:
            self.__trigger_event__(self.Heartbeat)
            number_of_beats -= 1

    def count_heartbeats(self):
        return self._count_heartbeats


class AbstractExampleRepository(AbstractEntityRepository):
    pass


def create_new_example(foo="", a="", b=""):
    """
    Factory method for example entities.

    :rtype: Example
    """
    return Example.__create__(foo=foo, a=a, b=b)



from uuid import UUID

from eventsourcing.application.policies import PersistencePolicy
from eventsourcing.example.domainmodel import create_new_example
from eventsourcing.example.infrastructure import ExampleRepository
from eventsourcing.infrastructure.eventstore import EventStore
from eventsourcing.infrastructure.sequenceditem import StoredEvent
from eventsourcing.infrastructure.sequenceditemmapper import SequencedItemMapper
from eventsourcing.infrastructure.sqlalchemy.datastore import (
    SQLAlchemyDatastore,
    SQLAlchemySettings,
)
from eventsourcing.infrastructure.sqlalchemy.manager import SQLAlchemyRecordManager
from eventsourcing.infrastructure.sqlalchemy.records import Base, StoredEventRecord
from eventsourcing.tests.datastore_tests.base import AbstractDatastoreTestCase

# This test replaces the default SequencedItem class with a StoredEvent class.
# How easy is it to customize the infrastructure to support that? We just need
# to define the new sequenced item class, define a suitable record class,
# and configure the other components. It's easy.


class ExampleApplicationWithAlternativeSequencedItemType(object):
    def __init__(self, session):
        self.event_store = EventStore(
            record_manager=SQLAlchemyRecordManager(
                session=session,
                record_class=StoredEventRecord,
                sequenced_item_class=StoredEvent,
            ),
            event_mapper=SequencedItemMapper(
                sequenced_item_class=StoredEvent,
                sequence_id_attr_name="originator_id",
                position_attr_name="originator_version",
            ),
        )
        self.repository = ExampleRepository(event_store=self.event_store)
        self.persistence_policy = PersistencePolicy(self.event_store)

    def close(self):
        self.persistence_policy.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()


class TestExampleWithAlternativeSequencedItemType(AbstractDatastoreTestCase):
    def setUp(self):
        super(TestExampleWithAlternativeSequencedItemType, self).setUp()
        self.datastore.setup_connection()
        self.datastore.drop_tables()  # something isn't dropping tables...
        self.datastore.setup_tables()

    def tearDown(self):
        self.datastore.drop_tables()
        self.datastore.close_connection()
        super(TestExampleWithAlternativeSequencedItemType, self).setUp()

    def construct_datastore(self):
        return SQLAlchemyDatastore(
            base=Base, settings=SQLAlchemySettings(), tables=(StoredEventRecord,)
        )

    def _test(self):
        with ExampleApplicationWithAlternativeSequencedItemType(
            self.datastore.session
        ) as app:
            # Create entity.
            entity1 = create_new_example(a="a", b="b")
            self.assertIsInstance(entity1.id, UUID)
            self.assertEqual(entity1.a, "a")
            self.assertEqual(entity1.b, "b")

            # Check there is a stored event.
            all_records = list(
                app.event_store.record_manager.get_notification_records()
            )
            self.assertEqual(1, len(all_records))
            stored_event = all_records[0]
            self.assertEqual(stored_event.originator_id, entity1.id)
            self.assertEqual(stored_event.originator_version, 0)

            # Read entity from repo.
            retrieved_obj = app.repository[entity1.id]
            self.assertEqual(retrieved_obj.id, entity1.id)



import datetime
from base64 import b64decode, b64encode
from collections import deque
from decimal import Decimal
from enum import Enum
from functools import singledispatch, wraps
from inspect import isfunction
from json import JSONDecoder, JSONEncoder
from types import FunctionType, MethodType
from uuid import UUID

import dateutil.parser

from eventsourcing.exceptions import EncoderTypeError
from eventsourcing.utils.topic import get_topic, resolve_topic

JSON_SEPARATORS = (",", ":")


def encoderpolicy(arg=None):
    """
    Decorator for encoder policy.

    Allows default behaviour to be built up from methods
    registered for different types of things, rather than
    chain of isinstance() calls in a long if-else block.
    """

    def _mutator(func):
        wrapped = singledispatch(func)

        @wraps(wrapped)
        def wrapper(*args, **kwargs):
            obj = kwargs.get("obj") or args[-1]
            return wrapped.dispatch(type(obj))(*args, **kwargs)

        wrapper.register = wrapped.register

        return wrapper

    assert isfunction(arg), arg
    return _mutator(arg)


def decoderpolicy(arg=None):
    """
    Decorator for decoder policy.

    Allows default behaviour to be built up from methods
    registered for different named keys, rather than
    chain of "in dict" queries in a long if-else block.
    """

    def _mutator(func):
        wrapped = func

        decoder_map = {}

        @wraps(wrapped)
        def wrapper(*args, **kwargs):

            d = kwargs.get("d") or args[-1]
            keys = list(d.keys())

            if len(keys) == 1:
                try:
                    decoder_func = decoder_map[keys[0]]
                except KeyError:
                    return d
                else:
                    return decoder_func(d)
            else:
                return d

        def register(key):
            def decorator(decoder_func):
                decoder_map[key] = decoder_func
                return decoder_func

            return decorator

        wrapper.register = register

        return wrapper

    assert isfunction(arg), arg
    return _mutator(arg)


class ObjectJSONEncoder(JSONEncoder):
    def encode(self, o):
        return super(ObjectJSONEncoder, self).encode(o).encode("utf8")

    def iterencode(self, o, _one_shot=False):
        if isinstance(o, tuple):
            o = {"__tuple__": {"topic": (get_topic(o.__class__)), "state": (list(o))}}
        return super(ObjectJSONEncoder, self).iterencode(o, _one_shot=_one_shot)

    def default(self, obj):
        try:
            return encoder(obj)
        except EncoderTypeError:
            return JSONEncoder.default(self, obj)


@encoderpolicy
def encoder(obj):
    """Extensible encoder function."""


class ObjectJSONDecoder(JSONDecoder):
    def __init__(self, object_hook=None, **kwargs):
        super(ObjectJSONDecoder, self).__init__(
            object_hook=object_hook or decoder, **kwargs
        )


@decoderpolicy
def decoder(d):
    """Extensible decoder function."""


@encoder.register(bytes)
def encode_bytes(obj):
    return {"__bytes__": b64str_from_bytes(obj)}


@encoder.register(type)
def encode_type(obj):
    return {"__type__": get_topic(obj)}


@encoder.register(MethodType)
def encode_method(obj):
    raise EncoderTypeError(obj)


@encoder.register(FunctionType)
def encode_function(obj):
    raise EncoderTypeError(obj)


@decoder.register("__type__")
def decode_type(d):
    return resolve_topic(d["__type__"])


@encoder.register(object)
def encode_object(obj):
    if hasattr(obj, "__slots__"):
        topic = get_topic(obj.__class__)
        state = {k: getattr(obj, k) for k in obj.__slots__}
        return {"__class__": {"topic": topic, "state": state}}
    elif hasattr(obj, "__dict__"):
        topic = get_topic(obj.__class__)
        state = obj.__dict__.copy()
        return {"__class__": {"topic": topic, "state": state}}
    else:
        raise EncoderTypeError(obj)


@decoder.register("__class__")
def decode_object(d):
    topic = d["__class__"]["topic"]
    state = d["__class__"]["state"]
    obj_class = resolve_topic(topic)
    obj = object.__new__(obj_class)
    if hasattr(obj, "__dict__"):
        obj.__dict__.update(state)
    else:
        for k, v in state.items():
            object.__setattr__(obj, k, v)
    return obj


@encoder.register(UUID)
def encode_uuid(obj):
    return {"UUID": obj.hex}


@decoder.register("UUID")
def decode_uuid(d):
    return UUID(d["UUID"])


@encoder.register(datetime.datetime)
def encode_datetime(obj):
    return {"ISO8601_datetime": obj.strftime("%Y-%m-%dT%H:%M:%S.%f%z")}


@decoder.register("ISO8601_datetime")
def decode_datetime(d):
    return dateutil.parser.parse(d["ISO8601_datetime"])


@encoder.register(datetime.date)
def encode_date(obj):
    return {"ISO8601_date": obj.isoformat()}


@decoder.register("ISO8601_date")
def decode_date(d):
    return datetime.datetime.strptime(d["ISO8601_date"], "%Y-%m-%d").date()


@encoder.register(datetime.time)
def encode_time(obj):
    return {"ISO8601_time": obj.strftime("%H:%M:%S.%f")}


@decoder.register("ISO8601_time")
def decode_time(d):
    hour, minute, seconds = d["ISO8601_time"].split(":")
    second, microsecond = seconds.split(".")
    return datetime.time(int(hour), int(minute), int(second), int(microsecond))


@encoder.register(Decimal)
def encode_decimal(obj):
    return {"__decimal__": str(obj)}


@decoder.register("__decimal__")
def decode_decimal(d):
    return Decimal(d["__decimal__"])


@encoder.register(Enum)
def encode_enum(obj):
    return {"__enum__": {"topic": get_topic(type(obj)), "name": obj.name}}


@decoder.register("__enum__")
def decode_enum(d):
    topic = d["__enum__"]["topic"]
    name = d["__enum__"]["name"]
    enum = resolve_topic(topic)
    return getattr(enum, name)


@encoder.register(deque)
def encode_deque(obj):
    return {"__deque__": {"topic": get_topic(type(obj)), "values": list(obj)}}


@decoder.register("__deque__")
def decode_deque(d):
    topic = d["__deque__"]["topic"]
    values = d["__deque__"]["values"]
    deque = resolve_topic(topic)
    return deque(values)


@decoder.register("__tuple__")
def decode_tuple(d):
    topic = d["__tuple__"]["topic"]
    state = d["__tuple__"]["state"]
    tuple_type = resolve_topic(topic)
    if topic == "builtins#tuple":
        # For standard tuple objects.
        obj = tuple_type(state)
    else:
        # For NamedTuple objects.
        obj = tuple_type(*state)
    return obj


@encoder.register(set)
def encode_set(obj):
    return {"__set__": sorted(list(obj))}


@decoder.register("__set__")
def decode_set(d):
    return set(d["__set__"])


@decoder.register("__bytes__")
def decode_bytes(d):
    return bytes_from_b64str(d["__bytes__"])


def b64str_from_bytes(value: bytes) -> str:
    return b64encode(value).decode("utf8")


def bytes_from_b64str(value):
    return b64decode(value.encode("utf8"))



from eventsourcing.tests.example_application_tests import base
from eventsourcing.tests.sequenced_item_tests.test_cassandra_record_manager import (
    WithCassandraRecordManagers,
)


class TestExampleApplicationWithCassandra(
    WithCassandraRecordManagers, base.ExampleApplicationTestCase
):
    pass



from typing import Iterator, NamedTuple, Tuple, Type
from uuid import UUID


class SequencedItem(NamedTuple):
    sequence_id: UUID
    position: int
    topic: str
    state: bytes


class StoredEvent(NamedTuple):
    originator_id: UUID
    originator_version: int
    topic: str
    state: bytes


class SequencedItemFieldNames(object):
    def __init__(self, sequenced_item_class: Type[NamedTuple]):
        self._field_names = sequenced_item_class._fields

    @property
    def sequence_id(self) -> str:
        # Sequence ID is assumed to be the first field of a sequenced item.
        return self._field_names[0]

    @property
    def position(self) -> str:
        # Position is assumed to be the second field of a sequenced item.
        return self._field_names[1]

    @property
    def topic(self) -> str:
        # Topic is assumed to be the third field of a sequenced item.
        return self._field_names[2]

    @property
    def state(self) -> str:
        # State is assumed to be the fourth field of a sequenced item.
        return self._field_names[3]

    @property
    def other_names(self) -> Tuple[str, ...]:
        return self._field_names[4:]

    def __iter__(self) -> Iterator[str]:
        for i in range(len(self._field_names)):
            yield self._field_names[i]



from abc import ABC, abstractmethod
from copy import deepcopy
from typing import Optional
from uuid import UUID

from eventsourcing.domain.model.events import AbstractSnapshot
from eventsourcing.domain.model.snapshot import Snapshot
from eventsourcing.infrastructure.base import AbstractEventStore, AbstractRecordManager
from eventsourcing.utils.topic import get_topic


class AbstractSnapshotStrategy(ABC):
    @abstractmethod
    def get_snapshot(
        self, entity_id: UUID, lt: Optional[int] = None, lte: Optional[int] = None
    ) -> Optional[AbstractSnapshot]:
        """
        Gets the last snapshot for entity, optionally until a particular version number.

        :rtype: Snapshot
        """

    @abstractmethod
    def take_snapshot(
        self, entity_id: UUID, entity: object, last_event_version: int
    ) -> Snapshot:
        """
        Takes a snapshot of entity, using given ID, state and version number.
        """


class EventSourcedSnapshotStrategy(AbstractSnapshotStrategy):
    """Snapshot strategy that uses an event sourced snapshot.
    """

    def __init__(
        self,
        snapshot_store: AbstractEventStore[AbstractSnapshot, AbstractRecordManager],
    ):
        self.snapshot_store = snapshot_store

    def get_snapshot(
        self, entity_id: UUID, lt: Optional[int] = None, lte: Optional[int] = None
    ) -> Optional[AbstractSnapshot]:
        """
        Gets the last snapshot for entity, optionally until a particular version number.

        :rtype: Snapshot
        """
        snapshots = self.snapshot_store.list_events(
            entity_id, lt=lt, lte=lte, limit=1, is_ascending=False
        )
        snapshot = None
        if len(snapshots) == 1:
            snapshot = snapshots[0]
        return snapshot

    # Todo: Rename as create_snapshot?
    def take_snapshot(
        self, entity_id: UUID, entity: object, last_event_version: int
    ) -> Snapshot:
        """
        Creates a Snapshot from the given state, and appends it
        to the snapshot store.

        :rtype: Snapshot
        """

        # Create the snapshot.
        snapshot = Snapshot(
            originator_id=entity_id,
            originator_version=last_event_version,
            topic=get_topic(entity.__class__),
            state=None if entity is None else deepcopy(entity.__dict__),
        )

        self.snapshot_store.store_events([snapshot])

        # Return the snapshot.
        return snapshot



from uuid import uuid4

from eventsourcing.example.domainmodel import Example
from eventsourcing.infrastructure.eventstore import EventStore
from eventsourcing.infrastructure.sequenceditem import SequencedItem
from eventsourcing.infrastructure.sequenceditemmapper import SequencedItemMapper
from eventsourcing.tests.datastore_tests.test_sqlalchemy import (
    SQLAlchemyDatastoreTestCase,
)
from eventsourcing.utils.topic import get_topic


class TestEventStore(SQLAlchemyDatastoreTestCase):
    def setUp(self):
        super(TestEventStore, self).setUp()
        if self.datastore is not None:
            self.datastore.setup_connection()
            self.datastore.setup_tables()

    def tearDown(self):
        if self.datastore is not None:
            self.datastore.drop_tables()
            self.datastore.close_connection()
        super(TestEventStore, self).tearDown()

    def construct_event_store(self):
        event_store = EventStore(
            record_manager=self.factory.construct_integer_sequenced_record_manager(),
            event_mapper=SequencedItemMapper(
                sequenced_item_class=SequencedItem,
                sequence_id_attr_name="originator_id",
                position_attr_name="originator_version",
            ),
        )
        return event_store

    def test_get_domain_events_deprecated(self):
        event_store = self.construct_event_store()

        # Check there are zero stored events in the repo.
        entity_id1 = uuid4()
        entity_events = event_store.get_domain_events(originator_id=entity_id1)
        self.assertEqual(0, len(list(entity_events)))

        # Store a domain event.
        event1 = Example.Created(
            a=1, b=2, originator_id=entity_id1, originator_topic=get_topic(Example)
        )
        event_store.store_events([event1])

        # Check there is one event in the event store.
        entity_events = event_store.get_domain_events(originator_id=entity_id1)
        self.assertEqual(1, len(list(entity_events)))

    def test_list_events(self):
        event_store = self.construct_event_store()

        # Check there are zero stored events in the repo.
        entity_id1 = uuid4()
        entity_events = event_store.list_events(originator_id=entity_id1)
        self.assertEqual(0, len(entity_events))

        # Check there are zero events in the event store, using iterator.
        entity_events = event_store.list_events(originator_id=entity_id1, page_size=1)
        self.assertEqual(0, len(entity_events))

        # Store a domain event.
        event1 = Example.Created(
            a=1, b=2, originator_id=entity_id1, originator_topic=get_topic(Example)
        )
        event_store.store_events([event1])

        # Check there is one event in the event store.
        entity_events = event_store.list_events(originator_id=entity_id1)
        self.assertEqual(1, len(entity_events))

        # Check there are two events in the event store, using iterator.
        entity_events = event_store.list_events(originator_id=entity_id1, page_size=1)
        self.assertEqual(1, len(entity_events))

        # Store another domain event.
        event1 = Example.AttributeChanged(
            a=1, b=2, originator_id=entity_id1, originator_version=1
        )
        event_store.store_events([event1])

        # Check there are two events in the event store.
        entity_events = event_store.list_events(originator_id=entity_id1)
        self.assertEqual(2, len(entity_events))

        # Check there are two events in the event store, using iterator.
        entity_events = event_store.list_events(originator_id=entity_id1, page_size=1)
        self.assertEqual(2, len(entity_events))

    def test_get_most_recent_event(self):
        event_store = self.construct_event_store()

        # Check there is no most recent event.
        entity_id1 = uuid4()
        entity_event = event_store.get_most_recent_event(originator_id=entity_id1)
        self.assertEqual(entity_event, None)

        # Store a domain event.
        event1 = Example.Created(
            a=1, b=2, originator_id=entity_id1, originator_topic=get_topic(Example)
        )
        event_store.store_events([event1])

        # Check there is an event.
        entity_event = event_store.get_most_recent_event(originator_id=entity_id1)
        self.assertEqual(entity_event, event1)

    def test_all_domain_events(self):
        event_store = self.construct_event_store()

        # Check there are zero domain events in total.
        domain_events = event_store.all_events()
        domain_events = list(domain_events)
        self.assertEqual(len(domain_events), 0)

        # Store a domain event.
        entity_id1 = uuid4()
        event1 = Example.Created(
            a=1, b=2, originator_id=entity_id1, originator_topic=get_topic(Example)
        )
        event_store.store_events([event1])

        # Store another domain event for the same entity.
        event1 = Example.AttributeChanged(
            a=1,
            b=2,
            originator_id=entity_id1,
            originator_version=1,
            __previous_hash__=event1.__event_hash__,
        )
        event_store.store_events([event1])

        # Store a domain event for a different entity.
        entity_id2 = uuid4()
        event1 = Example.Created(
            originator_topic=get_topic(Example), originator_id=entity_id2, a=1, b=2
        )
        event_store.store_events([event1])

        # Check there are three domain events in total.
        domain_events = event_store.all_events()
        domain_events = list(domain_events)
        self.assertEqual(len(domain_events), 3)



from typing import Any, NamedTuple, Optional, Type

from eventsourcing.infrastructure.axonserver.datastore import (
    AxonDatastore,
    AxonSettings,
)
from eventsourcing.infrastructure.axonserver.manager import AxonRecordManager
from eventsourcing.infrastructure.base import AbstractRecordManager
from eventsourcing.infrastructure.datastore import AbstractDatastore
from eventsourcing.infrastructure.factory import InfrastructureFactory


class AxonInfrastructureFactory(InfrastructureFactory):
    """
    Infrastructure factory for Axon infrastructure.
    """

    record_manager_class = AxonRecordManager
    integer_sequenced_record_class = None
    timestamp_sequenced_record_class = None
    snapshot_record_class = None
    tracking_record_class = None

    def __init__(
        self, axon_client=None, uri: Optional[str] = None, *args: Any, **kwargs: Any
    ):
        super(AxonInfrastructureFactory, self).__init__(*args, **kwargs)
        self.axon_client = axon_client
        self.uri = uri

    def construct_integer_sequenced_record_manager(
        self, **kwargs: Any
    ) -> AbstractRecordManager:
        """
        Constructs Axon record manager.

        :return: An Axon record manager.
        :rtype: AxonRecordManager
        """
        return super(
            AxonInfrastructureFactory, self
        ).construct_integer_sequenced_record_manager(**kwargs)

    def construct_record_manager(
        self,
        record_class: Optional[type],
        sequenced_item_class: Optional[Type[NamedTuple]] = None,
        **kwargs: Any
    ) -> AbstractRecordManager:
        """
        Constructs Axon record manager.

        :return: An Axon record manager.
        :rtype: AxonRecordManager
        """
        return super(AxonInfrastructureFactory, self).construct_record_manager(
            record_class,
            sequenced_item_class=sequenced_item_class,
            axon_client=self.axon_client,
            **kwargs
        )

    def construct_datastore(self) -> Optional[AbstractDatastore]:
        """
        Constructs Axon datastore.

        :rtype: AxonDatastore
        """
        datastore = AxonDatastore(settings=AxonSettings(uri=self.uri))
        assert datastore.axon_client
        if self.axon_client is None:
            self.axon_client = datastore.axon_client
        return datastore



from typing import Any, Optional

from eventsourcing.application.simple import ApplicationWithConcreteInfrastructure
from eventsourcing.infrastructure.factory import InfrastructureFactory
from eventsourcing.infrastructure.sqlalchemy.factory import (
    SQLAlchemyInfrastructureFactory,
)
from eventsourcing.infrastructure.sqlalchemy.records import (
    EntitySnapshotRecord,
    StoredEventRecord,
)


class SQLAlchemyApplication(ApplicationWithConcreteInfrastructure):
    infrastructure_factory_class = SQLAlchemyInfrastructureFactory
    stored_event_record_class = StoredEventRecord
    snapshot_record_class = EntitySnapshotRecord
    is_constructed_with_session = True
    tracking_record_class: Any = None

    def __init__(
        self,
        uri: Optional[str] = None,
        session: Optional[Any] = None,
        tracking_record_class: Any = None,
        **kwargs: Any
    ):
        self.uri = uri
        self._session = session
        self.tracking_record_class = (
            tracking_record_class or type(self).tracking_record_class
        )
        super(SQLAlchemyApplication, self).__init__(**kwargs)

    @property
    def session(self) -> Optional[Any]:
        return self._session

    def construct_infrastructure(self, *args: Any, **kwargs: Any) -> None:
        super(SQLAlchemyApplication, self).construct_infrastructure(
            session=self.session,
            uri=self.uri,
            tracking_record_class=self.tracking_record_class,
            *args,
            **kwargs
        )

    def construct_infrastructure_factory(
        self, *args: Any, **kwargs: Any
    ) -> InfrastructureFactory:
        return super(SQLAlchemyApplication, self).construct_infrastructure_factory(
            *args, **kwargs
        )

    def construct_datastore(self) -> None:
        super(SQLAlchemyApplication, self).construct_datastore()
        assert self._datastore
        assert self._datastore.session
        if self._session is None:
            self._session = self._datastore.session



import uuid
from typing import Dict, Generic, Optional, cast
from unittest.case import TestCase
from uuid import UUID

from eventsourcing.application.policies import PersistencePolicy
from eventsourcing.domain.model.aggregate import AggregateRoot
from eventsourcing.domain.model.decorators import attribute, subclassevents
from eventsourcing.domain.model.events import DomainEvent, assert_event_handlers_empty
from eventsourcing.exceptions import EventHashError, HeadHashError
from eventsourcing.infrastructure.eventsourcedrepository import EventSourcedRepository
from eventsourcing.infrastructure.eventstore import EventStore
from eventsourcing.infrastructure.sequenceditemmapper import SequencedItemMapper
from eventsourcing.infrastructure.sqlalchemy.manager import SQLAlchemyRecordManager
from eventsourcing.infrastructure.sqlalchemy.records import IntegerSequencedNoIDRecord
from eventsourcing.tests.sequenced_item_tests.test_sqlalchemy_record_manager import (
    SQLAlchemyRecordManagerTestCase,
)
from eventsourcing.utils.topic import get_topic, resolve_topic
from eventsourcing.whitehead import TEntity


class TestAggregateRootEvent(TestCase):
    def test_validate_aggregate_events(self):
        event1 = AggregateRoot.Created(
            originator_version=0,
            originator_id="1",
            originator_topic=get_topic(AggregateRoot),
        )
        event1.__check_hash__()

        # Chain another event.
        event2 = AggregateRoot.AttributeChanged(
            originator_version=1,
            originator_id="1",
            __previous_hash__=event1.__event_hash__,
        )
        event2.__check_hash__()

        # Chain another event.
        event3 = AggregateRoot.AttributeChanged(
            originator_version=2,
            originator_id="1",
            __previous_hash__=event2.__event_hash__,
        )
        event3.__check_hash__()

    def test_event_hash_error(self):
        event1 = AggregateRoot.Created(
            originator_version=0,
            originator_id="1",
            originator_topic=get_topic(AggregateRoot),
        )
        event1.__check_hash__()

        # Break the hash.
        event1.__dict__["event_hash"] = "damage"
        with self.assertRaises(EventHashError):
            event1.__check_hash__()


class TestExampleAggregateRoot(SQLAlchemyRecordManagerTestCase):
    def setUp(self) -> None:
        assert_event_handlers_empty()
        super(TestExampleAggregateRoot, self).setUp()
        self.app: ExampleDDDApplication = ExampleDDDApplication(self.datastore)

    def tearDown(self) -> None:
        self.app.close()
        super(TestExampleAggregateRoot, self).tearDown()
        assert_event_handlers_empty()

    def test_example_aggregate_event_classes(self):
        self.assertIn("Event", ExampleAggregateRoot.__dict__)
        self.assertIn("Created", ExampleAggregateRoot.__dict__)
        self.assertIn("Discarded", ExampleAggregateRoot.__dict__)
        self.assertIn("AttributeChanged", ExampleAggregateRoot.__dict__)

        self.assertEqual(ExampleAggregateRoot.Event.__name__, "Event")
        self.assertEqual(
            ExampleAggregateRoot.Event.__qualname__, "ExampleAggregateRoot.Event"
        )
        topic = "eventsourcing.tests.core_tests.test_aggregate_root#ExampleAggregateRoot.Event"
        self.assertEqual(get_topic(ExampleAggregateRoot.Event), topic)
        self.assertEqual(resolve_topic(topic), ExampleAggregateRoot.Event)

        self.assertEqual(ExampleAggregateRoot.Created.__name__, "Created")
        self.assertEqual(
            ExampleAggregateRoot.Created.__qualname__, "ExampleAggregateRoot.Created"
        )
        topic = "eventsourcing.tests.core_tests.test_aggregate_root#ExampleAggregateRoot.Created"
        self.assertEqual(get_topic(ExampleAggregateRoot.Created), topic)
        self.assertEqual(resolve_topic(topic), ExampleAggregateRoot.Created)
        self.assertTrue(
            issubclass(ExampleAggregateRoot.Created, ExampleAggregateRoot.Event)
        )

        self.assertEqual(ExampleAggregateRoot.Discarded.__name__, "Discarded")
        self.assertEqual(
            ExampleAggregateRoot.Discarded.__qualname__,
            "ExampleAggregateRoot.Discarded",
        )
        topic = "eventsourcing.tests.core_tests.test_aggregate_root#ExampleAggregateRoot.Discarded"
        self.assertEqual(get_topic(ExampleAggregateRoot.Discarded), topic)
        self.assertEqual(resolve_topic(topic), ExampleAggregateRoot.Discarded)
        self.assertTrue(
            issubclass(ExampleAggregateRoot.Discarded, ExampleAggregateRoot.Event)
        )

        self.assertEqual(ExampleAggregateRoot.ExampleCreated.__name__, "ExampleCreated")
        self.assertEqual(
            ExampleAggregateRoot.ExampleCreated.__qualname__,
            "ExampleAggregateRoot.ExampleCreated",
        )
        topic = "eventsourcing.tests.core_tests.test_aggregate_root#ExampleAggregateRoot.ExampleCreated"
        self.assertEqual(get_topic(ExampleAggregateRoot.ExampleCreated), topic)
        self.assertEqual(resolve_topic(topic), ExampleAggregateRoot.ExampleCreated)
        self.assertTrue(
            issubclass(ExampleAggregateRoot.ExampleCreated, ExampleAggregateRoot.Event)
        )

    def test_aggregate1_event_classes(self):
        self.assertIn("Event", Aggregate1.__dict__)
        self.assertIn("Created", Aggregate1.__dict__)
        self.assertIn("Discarded", Aggregate1.__dict__)
        self.assertIn("AttributeChanged", Aggregate1.__dict__)

        self.assertEqual(Aggregate1.Event.__name__, "Event")
        self.assertEqual(Aggregate1.Event.__qualname__, "Aggregate1.Event")
        topic = "eventsourcing.tests.core_tests.test_aggregate_root#Aggregate1.Event"
        self.assertEqual(get_topic(Aggregate1.Event), topic)
        self.assertEqual(resolve_topic(topic), Aggregate1.Event)

        self.assertEqual(Aggregate1.Created.__name__, "Created")
        self.assertEqual(Aggregate1.Created.__qualname__, "Aggregate1.Created")
        topic = "eventsourcing.tests.core_tests.test_aggregate_root#Aggregate1.Created"
        self.assertEqual(get_topic(Aggregate1.Created), topic)
        self.assertEqual(resolve_topic(topic), Aggregate1.Created)
        self.assertTrue(issubclass(Aggregate1.Created, Aggregate1.Event))

        self.assertEqual(Aggregate1.Discarded.__name__, "Discarded")
        self.assertEqual(Aggregate1.Discarded.__qualname__, "Aggregate1.Discarded")
        topic = (
            "eventsourcing.tests.core_tests.test_aggregate_root#Aggregate1" ".Discarded"
        )
        self.assertEqual(get_topic(Aggregate1.Discarded), topic)
        self.assertEqual(resolve_topic(topic), Aggregate1.Discarded)
        self.assertTrue(issubclass(Aggregate1.Discarded, Aggregate1.Event))

        self.assertEqual(Aggregate1.ExampleCreated.__name__, "ExampleCreated")
        self.assertEqual(
            Aggregate1.ExampleCreated.__qualname__, "Aggregate1.ExampleCreated"
        )
        topic = (
            "eventsourcing.tests.core_tests.test_aggregate_root#Aggregate1"
            ".ExampleCreated"
        )
        self.assertEqual(get_topic(Aggregate1.ExampleCreated), topic)
        self.assertEqual(resolve_topic(topic), Aggregate1.ExampleCreated)
        self.assertTrue(issubclass(Aggregate1.ExampleCreated, Aggregate1.Event))
        self.assertTrue(issubclass(Aggregate1.SomethingElseOccurred, Aggregate1.Event))

    def test_aggregate2_event_classes(self):
        self.assertIn("Event", Aggregate2.__dict__)
        self.assertIn("Created", Aggregate2.__dict__)
        self.assertIn("Discarded", Aggregate2.__dict__)
        self.assertIn("AttributeChanged", Aggregate2.__dict__)

        self.assertEqual(Aggregate2.Event.__name__, "Event")
        self.assertEqual(Aggregate2.Event.__qualname__, "Aggregate2.Event")
        topic = "eventsourcing.tests.core_tests.test_aggregate_root#Aggregate2.Event"
        self.assertEqual(get_topic(Aggregate2.Event), topic)
        self.assertEqual(resolve_topic(topic), Aggregate2.Event)

        self.assertEqual(Aggregate2.Created.__name__, "Created")
        self.assertEqual(Aggregate2.Created.__qualname__, "Aggregate2.Created")
        topic = "eventsourcing.tests.core_tests.test_aggregate_root#Aggregate2.Created"
        self.assertEqual(get_topic(Aggregate2.Created), topic)
        self.assertEqual(resolve_topic(topic), Aggregate2.Created)
        self.assertTrue(issubclass(Aggregate2.Created, Aggregate2.Event))

        self.assertEqual(Aggregate2.Discarded.__name__, "Discarded")
        self.assertEqual(Aggregate2.Discarded.__qualname__, "Aggregate2.Discarded")
        topic = (
            "eventsourcing.tests.core_tests.test_aggregate_root#Aggregate2" ".Discarded"
        )
        self.assertEqual(get_topic(Aggregate2.Discarded), topic)
        self.assertEqual(resolve_topic(topic), Aggregate2.Discarded)
        self.assertTrue(issubclass(Aggregate2.Discarded, Aggregate2.Event))

        self.assertEqual(Aggregate2.ExampleCreated.__name__, "ExampleCreated")
        self.assertEqual(
            Aggregate2.ExampleCreated.__qualname__, "Aggregate2.ExampleCreated"
        )
        topic = (
            "eventsourcing.tests.core_tests.test_aggregate_root#Aggregate2"
            ".ExampleCreated"
        )
        self.assertEqual(get_topic(Aggregate2.ExampleCreated), topic)
        self.assertEqual(resolve_topic(topic), Aggregate2.ExampleCreated)
        self.assertTrue(issubclass(Aggregate2.ExampleCreated, Aggregate2.Event))

    def test_aggregate3_lifecycle(self):
        # type: () -> None

        # Create a new aggregate.
        aggregate = self.app.create_aggregate3()

        self.assertIsInstance(aggregate, Aggregate3)

    def test_aggregate1_lifecycle(self):
        # type: () -> None

        # Create a new aggregate.
        aggregate = self.app.create_aggregate1()

        self.assertIsInstance(aggregate, Aggregate1)

        # Check it's got a head hash.
        self.assertTrue(aggregate.__head__)
        last_next_hash = aggregate.__head__

        # Check it does not exist in the repository.
        self.assertNotIn(aggregate.id, self.app.repository)

        # Save the aggregate.
        aggregate.__save__()

        # Check it now exists in the repository.
        self.assertIn(aggregate.id, self.app.repository)

        # Change an attribute of the aggregate root entity.
        self.assertNotEqual(aggregate.foo, "bar")
        aggregate.foo = "bar"
        self.assertEqual(aggregate.foo, "bar")

        # Check the head hash has changed.
        self.assertNotEqual(aggregate.__head__, last_next_hash)
        last_next_hash = aggregate.__head__

        self.assertIn(aggregate.id, self.app.repository)

        self.assertNotEqual(self.app.repository[aggregate.id].foo, "bar")
        aggregate.__save__()
        self.assertEqual(self.app.repository[aggregate.id].foo, "bar")

        # Check the aggregate has zero entities.
        self.assertEqual(aggregate.count_examples(), 0)

        # Check the aggregate has zero entities.
        self.assertEqual(aggregate.count_examples(), 0)

        # Ask the aggregate to create an entity within itself.
        aggregate.create_new_example()

        # Check the aggregate has one entity.
        self.assertEqual(aggregate.count_examples(), 1)

        # Check the aggregate in the repo still has zero entities.
        self.assertEqual(
            self.app.repository[aggregate.id].count_examples(), 0
        )

        # Check the head hash has changed.
        self.assertNotEqual(aggregate.__head__, last_next_hash)
        last_next_hash = aggregate.__head__

        # Call save().
        aggregate.__save__()

        # Check the aggregate in the repo now has one entity.
        self.assertEqual(
            self.app.repository[aggregate.id].count_examples(), 1
        )

        # Create two more entities within the aggregate.
        aggregate.create_new_example()
        aggregate.create_new_example()

        # Save both "entity created" events in one atomic transaction.
        aggregate.__save__()

        # Check the aggregate in the repo now has three entities.
        self.assertEqual(
            self.app.repository[aggregate.id].count_examples(), 3
        )

        # Discard the aggregate, calls save().
        aggregate.__discard__()
        aggregate.__save__()

        # Check the next hash has changed.
        self.assertNotEqual(aggregate.__head__, last_next_hash)

        # Check the aggregate no longer exists in the repo.
        self.assertNotIn(aggregate.id, self.app.repository)

    def test_both_types(self):
        # Create a new aggregate.
        aggregate1 = self.app.create_aggregate1()
        aggregate2 = self.app.create_aggregate2()

        aggregate1.__save__()
        aggregate2.__save__()

        self.assertIsInstance(aggregate1, Aggregate1)
        self.assertIsInstance(aggregate2, Aggregate2)

        self.assertEqual(aggregate1.foo, "")
        self.assertEqual(aggregate2.foo, "")

        aggregate1.foo = "bar"
        aggregate2.foo = "baz"

        aggregate1.__save__()
        aggregate2.__save__()

        aggregate1 = self.app.repository[aggregate1.id]
        aggregate2 = self.app.repository[aggregate2.id]

        self.assertIsInstance(aggregate1, Aggregate1)
        self.assertIsInstance(aggregate2, Aggregate2)

        self.assertEqual(aggregate1.foo, "bar")
        self.assertEqual(aggregate2.foo, "baz")

        aggregate1.__discard__()
        aggregate1.__save__()
        self.assertFalse(aggregate1.id in self.app.repository)
        self.assertTrue(aggregate2.id in self.app.repository)

    def test_validate_previous_hash_error(self):
        # Check event has valid originator head.
        aggregate = Aggregate1(id="1", foo="bar", __created_on__=0, __version__=0)
        event = Aggregate1.AttributeChanged(
            name="foo",
            value="bar",
            originator_id="1",
            originator_version=1,
            __previous_hash__=aggregate.__head__,
        )
        event.__check_obj__(aggregate)

        # Check OriginatorHeadError is raised if the originator head is wrong.
        event.__dict__["__previous_hash__"] += "damage"
        with self.assertRaises(HeadHashError):
            event.__check_obj__(aggregate)


@subclassevents
class ExampleAggregateRoot(AggregateRoot):
    def __init__(self, foo="", **kwargs):
        super(ExampleAggregateRoot, self).__init__(**kwargs)
        self._entities: Dict[UUID, Example] = {}
        self._foo = foo

    @attribute
    def foo(self):
        """Simple event sourced attribute called 'foo'."""

    def create_new_example(self):
        assert not self.__is_discarded__
        self.__trigger_event__(self.ExampleCreated, entity_id=uuid.uuid4())

    def count_examples(self):
        return len(self._entities)

    class ExampleCreated(DomainEvent[TEntity]):
        """Published when an example entity is created within the aggregate."""

        def __init__(self, entity_id, **kwargs):
            super().__init__(entity_id=entity_id, **kwargs)

        @property
        def entity_id(self):
            return self.__dict__["entity_id"]

        def __mutate__(self, obj: Optional[TEntity]) -> Optional[TEntity]:
            obj = super().__mutate__(obj)
            if obj:
                assert isinstance(obj, ExampleAggregateRoot)
                entity = Example(entity_id=self.entity_id)
                obj._entities[entity.id] = entity
            return obj

    class SomethingElseOccurred(DomainEvent):
        pass


@subclassevents
class Aggregate1(ExampleAggregateRoot):
    pass


class Aggregate2(ExampleAggregateRoot):
    __subclassevents__ = True


class Aggregate3(Aggregate2):
    pass


class AggregateRepository(EventSourcedRepository):
    pass


class Example(object):
    """
    Example domain entity.
    """

    def __init__(self, entity_id):
        self._id = entity_id

    @property
    def id(self):
        return self._id


class ExampleDDDApplication(Generic[TEntity]):
    def __init__(self, datastore):
        event_store = EventStore(
            record_manager=SQLAlchemyRecordManager(
                session=datastore.session, record_class=IntegerSequencedNoIDRecord
            ),
            event_mapper=SequencedItemMapper(
                sequence_id_attr_name="originator_id",
                position_attr_name="originator_version",
            ),
        )
        self.repository = EventSourcedRepository(event_store=event_store)
        self.persistence_policy = PersistencePolicy(
            persist_event_type=ExampleAggregateRoot.Event, event_store=event_store
        )

    def create_aggregate1(self) -> Aggregate1:
        """
        Factory method, creates and returns a new aggregate1 root entity.
        """
        a: Aggregate1 = Aggregate1.__create__()
        return a

    def create_aggregate2(self) -> Aggregate2:
        """
        Factory method, creates and returns a new aggregate1 root entity.
        """
        a: Aggregate2 = Aggregate2.__create__()
        return a

    def create_aggregate3(self) -> Aggregate3:
        """
        Factory method, creates and returns a new aggregate1 root entity.
        """
        a = Aggregate3.__create__()
        return a

    def close(self):
        self.persistence_policy.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()



import os
from abc import ABCMeta
from decimal import Decimal
from typing import Any, Dict, Optional, Sequence, Type, TypeVar
from uuid import UUID, uuid4

from eventsourcing.domain.model.decorators import subclassevents
from eventsourcing.domain.model.events import (
    AttributeChangedEvent,
    CreatedEvent,
    DiscardedEvent,
    DomainEvent,
    EventWithHash,
    EventWithOriginatorID,
    EventWithOriginatorVersion,
    EventWithTimestamp,
    publish,
)
from eventsourcing.domain.model.versioning import Upcastable
from eventsourcing.exceptions import (
    EntityIsDiscarded,
    HeadHashError,
    OriginatorIDError,
    OriginatorVersionError,
)
from eventsourcing.utils.times import decimaltimestamp_from_uuid
from eventsourcing.utils.topic import get_topic, resolve_topic
from eventsourcing.whitehead import EnduringObject

# Need to deal with the fact that Python3.6 had GenericMeta.
# Todo: Delete this try/except when dropping support for Python 3.6.
try:
    from typing import GenericMeta

    ABCMeta = GenericMeta  # type: ignore

except ImportError:
    pass


class MetaDomainEntity(ABCMeta):
    __subclassevents__ = False

    # Todo: Delete the '**kwargs' when library no longer supports Python3.6.
    #  - When we started using typing.Generic, we started getting
    #    an error in 3.6 (only) "unexpected keyword argument 'tvars'"
    #    which was cured by adding **kwargs here. It's not needed
    #    for Python3.7, and only supports backward compatibility.
    #    So it can be removed when support for Python 3.6 dropped.
    def __init__(cls, name: str, *args: Any, **kwargs: Any) -> None:
        super().__init__(name, *args, **kwargs)
        if name == "_gorg":
            # Todo: Also Remove this block when dropping support for Python 3.6.
            # Needed in 3.6 only, stops infinite recursion between typing and abc
            # doing subclass checks. Don't know why. Seems issue fixed in Python 3.7.
            pass
        elif cls.__subclassevents__ is True:
            # Redefine entity domain events.
            subclassevents(cls)


TDomainEntity = TypeVar("TDomainEntity", bound="DomainEntity")
TDomainEvent = TypeVar("TDomainEvent", bound="DomainEntity.Event")


class DomainEntity(Upcastable, EnduringObject, metaclass=MetaDomainEntity):
    """
    Supertype for domain model entity.
    """

    __subclassevents__ = False

    class Event(EventWithOriginatorID[TDomainEntity]):
        """
        Supertype for events of domain model entities.
        """

        def __check_obj__(self, obj: TDomainEntity) -> None:
            """
            Checks state of obj before mutating.

            :param obj: Domain entity to be checked.

            :raises OriginatorIDError: if the originator_id is mismatched
            """
            assert isinstance(obj, DomainEntity)  # For PyCharm navigation.
            # Assert ID matches originator ID.
            if obj.id != self.originator_id:
                raise OriginatorIDError(
                    "'{}' not equal to event originator ID '{}'"
                    "".format(obj.id, self.originator_id)
                )

    @classmethod
    def __create__(
        cls: Type[TDomainEntity],
        originator_id: Optional[UUID] = None,
        event_class: Optional[Type["DomainEntity.Created[TDomainEntity]"]] = None,
        **kwargs: Any,
    ) -> TDomainEntity:
        """
        Creates a new domain entity.

        Constructs a "created" event, constructs the entity object
        from the event, publishes the "created" event, and returns
        the new domain entity object.

        :param cls DomainEntity: Class of domain event
        :param originator_id: ID of the new domain entity (defaults to ``uuid4()``).
        :param event_class: Domain event class to be used for the "created" event.
        :param kwargs: Other named attribute values of the "created" event.
        :return: New domain entity object.
        :rtype: DomainEntity
        """

        if originator_id is None:
            originator_id = uuid4()

        if event_class is None:
            assert issubclass(cls, DomainEntity)  # For navigation in PyCharm.
            created_event_class: Type[DomainEntity.Created[TDomainEntity]] = cls.Created
        else:
            created_event_class = event_class

        event = created_event_class(
            originator_id=originator_id, originator_topic=get_topic(cls), **kwargs
        )

        obj = event.__mutate__(None)

        assert obj is not None, "{} returned None".format(
            type(event).__mutate__.__qualname__
        )

        obj.__publish__([event])
        return obj

    class Created(CreatedEvent[TDomainEntity], Event[TDomainEntity]):
        """
        Triggered when an entity is created.
        """

        def __init__(self, originator_topic: str, **kwargs: Any):
            super(DomainEntity.Created, self).__init__(
                originator_topic=originator_topic, **kwargs
            )

        @property
        def originator_topic(self) -> str:
            """
            Topic (a string) representing the class of the originating domain entity.

            :rtype: str
            """
            return self.__dict__["originator_topic"]

        def __mutate__(self, obj: Optional[TDomainEntity]) -> Optional[TDomainEntity]:
            """
            Constructs object from an entity class,
            which is obtained by resolving the originator topic,
            unless it is given as method argument ``entity_class``.

            :param entity_class: Class of domain entity to be constructed.
            """
            entity_class: Type[TDomainEntity] = resolve_topic(self.originator_topic)
            return entity_class(**self.__entity_kwargs__)

        @property
        def __entity_kwargs__(self) -> Dict[str, Any]:
            kwargs = self.__dict__.copy()
            kwargs["id"] = kwargs.pop("originator_id")
            kwargs.pop("originator_topic", None)
            kwargs.pop("__event_topic__", None)
            kwargs.pop("__event_hash_method_name__", None)
            return kwargs

    def __init__(self, id: UUID):
        self._id = id
        self.__is_discarded__ = False

    @property
    def id(self) -> UUID:
        """The immutable ID of the domain entity.

        This value is set using the ``originator_id`` of the
        "created" event constructed by ``__create__()``.

        An entity ID allows an instance to be
        referenced and distinguished from others, even
        though its state may change over time.

        This attribute has the normal "public" format for a Python object
        attribute name, because by definition all domain entities have an ID.
        """
        return self._id

    def __change_attribute__(
        self: TDomainEntity, name: str, value: Any, **kwargs
    ) -> None:
        """
        Changes named attribute with the given value,
        by triggering an AttributeChanged event.
        """
        event_class: Type[
            "DomainEntity.AttributeChanged[TDomainEntity]"
        ] = self.AttributeChanged
        assert isinstance(self, DomainEntity)  # For PyCharm navigation.
        self.__trigger_event__(
            event_class=event_class, name=name, value=value, **kwargs
        )

    class AttributeChanged(Event[TDomainEntity], AttributeChangedEvent[TDomainEntity]):
        """
        Triggered when a named attribute is assigned a new value.
        """

        def __mutate__(self, obj: Optional[TDomainEntity]) -> Optional[TDomainEntity]:
            obj = super(DomainEntity.AttributeChanged, self).__mutate__(obj)
            setattr(obj, self.name, self.value)
            return obj

    def __discard__(self: TDomainEntity, **kwargs) -> None:
        """
        Discards self, by triggering a Discarded event.
        """
        event_class: Type["DomainEntity.Discarded[TDomainEntity]"] = self.Discarded
        assert isinstance(self, DomainEntity)  # For PyCharm navigation.
        self.__trigger_event__(event_class=event_class, **kwargs)

    class Discarded(DiscardedEvent[TDomainEntity], Event[TDomainEntity]):
        """
        Triggered when a DomainEntity is discarded.
        """

        def __mutate__(self, obj: Optional[TDomainEntity]) -> Optional[TDomainEntity]:
            obj = super(DomainEntity.Discarded, self).__mutate__(obj)
            if obj is not None:
                assert isinstance(obj, DomainEntity)  # For PyCharm navigation.
                obj.__is_discarded__ = True
            return None

    def __assert_not_discarded__(self) -> None:
        """
        Asserts that this entity has not been discarded.

        Raises EntityIsDiscarded exception if entity has been discarded already.
        """
        if self.__is_discarded__:
            raise EntityIsDiscarded("Entity is discarded")

    def __trigger_event__(self, event_class: Type[TDomainEvent], **kwargs: Any) -> None:
        """
        Constructs, applies, and publishes a domain event.
        """
        self.__assert_not_discarded__()
        event: TDomainEvent = event_class(originator_id=self.id, **kwargs)
        self.__mutate__(event)
        self.__publish__([event])

    def __mutate__(self, event: TDomainEvent) -> None:
        """
        Mutates this entity with the given event.

        This method calls on the event object to mutate this
        entity, because the mutation behaviour of different types
        of events was usefully factored onto the event classes, and
        the event mutate() method is the most convenient way to
        defined behaviour in domain models.

        However, as an alternative to implementing the mutate()
        method on domain model events, this method can be extended
        with a method that is capable of mutating an entity for all
        the domain event classes introduced by the entity class.

        Similarly, this method can be overridden entirely in subclasses,
        so long as all of the mutation behaviour is implemented in the
        mutator function, including the mutation behaviour of the events
        defined on the library event classes that would no longer be invoked.

        However, if the entity class defines a mutator function, or if a
        separate mutator function is used, then it must be involved in
        the event sourced repository used to replay events, which by default
        knows nothing about the domain entity class. In practice, this
        means having a repository for each kind of entity, rather than
        the application just having one repository, with each repository
        having a mutator function that can project the entity events
        into an entity.
        """
        assert isinstance(event, DomainEntity.Event)
        event.__mutate__(self)

    def __publish__(self, event: Sequence[TDomainEvent]) -> None:
        """
        Publishes given event for subscribers in the application.

        :param event: domain event or list of events
        """
        self.__publish_to_subscribers__(event)

    def __publish_to_subscribers__(self, events: Sequence[TDomainEvent]) -> None:
        """
        Actually dispatches given event to publish-subscribe mechanism.

        :param events: list of domain events
        """
        publish(events)

    def __eq__(self, other: object) -> bool:
        return type(self) == type(other) and self.__dict__ == other.__dict__

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)


TEntityWithHashchain = TypeVar("TEntityWithHashchain", bound="EntityWithHashchain")

GENESIS_HASH: str = os.getenv("GENESIS_HASH", "")


class EntityWithHashchain(DomainEntity):
    __genesis_hash__ = GENESIS_HASH

    def __init__(self, *args: Any, **kwargs: Any):
        super(EntityWithHashchain, self).__init__(*args, **kwargs)
        self.__head__: str = type(self).__genesis_hash__

    class Event(
        EventWithHash[TEntityWithHashchain], DomainEntity.Event[TEntityWithHashchain]
    ):
        """
        Supertype for events of domain entities.
        """

        def __mutate__(
            self, obj: Optional[TEntityWithHashchain]
        ) -> Optional[TEntityWithHashchain]:
            # Call super method.
            obj = super(EntityWithHashchain.Event, self).__mutate__(obj)

            # Set entity head from event hash.
            #  - unless just discarded...
            if obj is not None:
                assert isinstance(obj, EntityWithHashchain)
                obj.__head__ = self.__event_hash__

            return obj

        def __check_obj__(self, obj: TEntityWithHashchain) -> None:
            """
            Extends superclass method by checking the __previous_hash__
            of this event matches the __head__ hash of the entity obj.
            """
            # Call super method.
            super(EntityWithHashchain.Event, self).__check_obj__(obj)
            assert isinstance(obj, EntityWithHashchain)  # For PyCharm navigation.
            # Assert __head__ equals __previous_hash__.
            if obj.__head__ != self.__dict__.get("__previous_hash__"):
                raise HeadHashError(obj.id, obj.__head__, type(self))

    class Created(
        Event[TEntityWithHashchain], DomainEntity.Created[TEntityWithHashchain]
    ):
        @property
        def __entity_kwargs__(self) -> Dict[str, Any]:
            # Get super property.
            kwargs = super(EntityWithHashchain.Created, self).__entity_kwargs__

            # Drop the event hashes.
            kwargs.pop("__event_hash__", None)
            kwargs.pop("__previous_hash__", None)

            return kwargs

    class AttributeChanged(
        Event[TEntityWithHashchain], DomainEntity.AttributeChanged[TEntityWithHashchain]
    ):
        pass

    class Discarded(
        Event[TEntityWithHashchain], DomainEntity.Discarded[TEntityWithHashchain]
    ):
        def __mutate__(
            self, obj: Optional[TEntityWithHashchain]
        ) -> Optional[TEntityWithHashchain]:
            # Set entity head from event hash.
            if obj:
                assert isinstance(obj, EntityWithHashchain)  # For PyCharm navigation.
                obj.__head__ = self.__event_hash__

            # Call super method.
            return super(EntityWithHashchain.Discarded, self).__mutate__(obj)

    @classmethod
    def __create__(
        cls: Type[TEntityWithHashchain],
        originator_id: Optional[UUID] = None,
        event_class: Optional[
            Type["DomainEntity.Created[TEntityWithHashchain]"]
        ] = None,
        **kwargs: Any,
    ) -> TEntityWithHashchain:
        # Initialise the hash-chain with "genesis hash".
        kwargs["__previous_hash__"] = getattr(cls, "__genesis_hash__", GENESIS_HASH)
        assert issubclass(cls, EntityWithHashchain)  # For PyCharm navigation.
        obj = super(EntityWithHashchain, cls).__create__(
            originator_id=originator_id, event_class=event_class, **kwargs
        )
        assert isinstance(obj, EntityWithHashchain)  # For PyCharm type checking.
        return obj

    def __trigger_event__(self, event_class: Type[TDomainEvent], **kwargs: Any) -> None:
        kwargs["__previous_hash__"] = self.__head__
        super(EntityWithHashchain, self).__trigger_event__(event_class, **kwargs)


TVersionedEntity = TypeVar("TVersionedEntity", bound="VersionedEntity")
TVersionedEvent = TypeVar("TVersionedEvent", bound="VersionedEntity.Event")


class VersionedEntity(DomainEntity):
    def __init__(self, __version__: int, **kwargs: Any):
        super().__init__(**kwargs)
        self.___version__: int = __version__

    @property
    def __version__(self) -> int:
        return self.___version__

    def __trigger_event__(self, event_class: Type[TDomainEvent], **kwargs: Any) -> None:
        """
        Increments the version number when an event is triggered.

        The event carries the version number that the originator
        will have when the originator is mutated with this event.
        (The event's "originator" version isn't the version of the
        originator before the event was triggered, but represents
        the result of the work of incrementing the version, which
        is then set in the event as normal. The Created event has
        version 0, and a newly created instance is at version 0.
        The second event has originator version 1, and so will the
        originator when the second event has been applied.
        """
        # Do the work of incrementing the version number.
        next_version = self.__version__ + 1
        # Trigger an event with the result of this work.
        super(VersionedEntity, self).__trigger_event__(
            event_class=event_class, originator_version=next_version, **kwargs
        )

    class Event(
        EventWithOriginatorVersion[TVersionedEntity],
        DomainEntity.Event[TVersionedEntity],
    ):
        """Supertype for events of versioned entities."""

        def __mutate__(
            self, obj: Optional[TVersionedEntity]
        ) -> Optional[TVersionedEntity]:
            obj = super(VersionedEntity.Event, self).__mutate__(obj)
            if obj is not None:
                assert isinstance(obj, VersionedEntity)  # For PyCharm navigation.
                obj.___version__ = self.originator_version
            return obj

        def __check_obj__(self, obj: TVersionedEntity) -> None:
            """
            Extends superclass method by checking the event's
            originator version follows (1 +) this entity's version.
            """
            super(VersionedEntity.Event, self).__check_obj__(obj)
            assert isinstance(obj, VersionedEntity)  # For PyCharm navigation.
            # Assert the version sequence is correct.
            if self.originator_version != obj.__version__ + 1:
                raise OriginatorVersionError(
                    (
                        "Event takes entity to version {}, "
                        "but entity is currently at version {}. "
                        "Event type: '{}', entity type: '{}', entity ID: '{}'"
                        "".format(
                            self.originator_version,
                            obj.__version__,
                            type(self).__name__,
                            type(obj).__name__,
                            obj._id,
                        )
                    )
                )

    class Created(DomainEntity.Created[TVersionedEntity], Event[TVersionedEntity]):
        """Published when a VersionedEntity is created."""

        def __init__(self, originator_version: int = 0, *args: Any, **kwargs: Any):
            super(VersionedEntity.Created, self).__init__(
                originator_version=originator_version, *args, **kwargs
            )

        @property
        def __entity_kwargs__(self) -> Dict[str, Any]:
            # Get super property.
            kwargs = super(VersionedEntity.Created, self).__entity_kwargs__
            kwargs["__version__"] = kwargs.pop("originator_version")
            return kwargs

    class AttributeChanged(
        Event[TVersionedEntity], DomainEntity.AttributeChanged[TVersionedEntity]
    ):
        """Published when a VersionedEntity is changed."""

    class Discarded(Event[TVersionedEntity], DomainEntity.Discarded[TVersionedEntity]):
        """Published when a VersionedEntity is discarded."""


class EntityWithECC(DomainEntity):
    """
    Entity whose events have event ID, correlation ID, and causation ID.
    """

    class Event(DomainEntity.Event):
        def __init__(self, *, processed_event=None, application_name, **kwargs):

            event_id = kwargs.get("event_id") or "{}:{}:{}".format(
                application_name, kwargs["originator_id"], kwargs["originator_version"]
            )
            kwargs["event_id"] = event_id
            if processed_event:
                kwargs["causation_id"] = processed_event.event_id
                kwargs["correlation_id"] = processed_event.correlation_id
            else:
                kwargs["causation_id"] = None
                kwargs["correlation_id"] = event_id

            super().__init__(**kwargs)

        @property
        def event_id(self):
            return self.__dict__["event_id"]

        @property
        def correlation_id(self):
            return self.__dict__["correlation_id"]

        @property
        def causation_id(self):
            return self.__dict__["causation_id"]

    class Created(DomainEntity.Created, Event):
        pass

    class AttributeChanged(Event, DomainEntity.AttributeChanged):
        pass

    class Discarded(Event, DomainEntity.Discarded):
        pass

    def __init__(self, *, event_id, correlation_id, causation_id, **kwargs):
        _ = event_id, correlation_id, causation_id  # accept, but ignore
        super(EntityWithECC, self).__init__(**kwargs)


TTimestampedEntity = TypeVar("TTimestampedEntity", bound="TimestampedEntity")


class TimestampedEntity(DomainEntity):
    def __init__(self, __created_on__: Decimal, **kwargs: Any):
        super(TimestampedEntity, self).__init__(**kwargs)
        self.___created_on__ = __created_on__
        self.___last_modified__ = __created_on__

    @property
    def __created_on__(self) -> Decimal:
        return self.___created_on__

    @property
    def __last_modified__(self) -> Decimal:
        return self.___last_modified__

    class Event(
        DomainEntity.Event[TTimestampedEntity], EventWithTimestamp[TTimestampedEntity]
    ):
        """Supertype for events of timestamped entities."""

        def __mutate__(
            self, obj: Optional[TTimestampedEntity]
        ) -> Optional[TTimestampedEntity]:
            """Updates 'obj' with values from self."""
            obj = super(TimestampedEntity.Event, self).__mutate__(obj)
            if obj is not None:
                assert isinstance(obj, TimestampedEntity)  # For PyCharm navigation.
                obj.___last_modified__ = self.timestamp
            return obj

    class Created(DomainEntity.Created[TTimestampedEntity], Event[TTimestampedEntity]):
        """Published when a TimestampedEntity is created."""

        @property
        def __entity_kwargs__(self) -> Dict[str, Any]:
            # Get super property.
            kwargs = super(TimestampedEntity.Created, self).__entity_kwargs__
            kwargs["__created_on__"] = kwargs.pop("timestamp")
            return kwargs

    class AttributeChanged(
        Event[TTimestampedEntity], DomainEntity.AttributeChanged[TTimestampedEntity]
    ):
        """Published when a TimestampedEntity is changed."""

    class Discarded(
        Event[TTimestampedEntity], DomainEntity.Discarded[TTimestampedEntity]
    ):
        """Published when a TimestampedEntity is discarded."""


# Todo: Move stuff from "test_customise_with_alternative_domain_event_type" in here (
#  to define event classes
#  and update ___last_event_id__ in mutate method).


class TimeuuidedEntity(DomainEntity):
    def __init__(self, event_id: UUID, **kwargs: Any) -> None:
        super(TimeuuidedEntity, self).__init__(**kwargs)
        self.___initial_event_id__ = event_id
        self.___last_event_id__ = event_id

    @property
    def __created_on__(self) -> Decimal:
        return decimaltimestamp_from_uuid(self.___initial_event_id__)

    @property
    def __last_modified__(self) -> Decimal:
        return decimaltimestamp_from_uuid(self.___last_event_id__)


TTimestampedVersionedEntity = TypeVar(
    "TTimestampedVersionedEntity", bound="TimestampedVersionedEntity"
)


class TimestampedVersionedEntity(TimestampedEntity, VersionedEntity):
    class Event(
        TimestampedEntity.Event[TTimestampedVersionedEntity],
        VersionedEntity.Event[TTimestampedVersionedEntity],
    ):
        """Supertype for events of timestamped, versioned entities."""

    class Created(
        TimestampedEntity.Created[TTimestampedVersionedEntity],
        VersionedEntity.Created,
        Event[TTimestampedVersionedEntity],
    ):
        """Published when a TimestampedVersionedEntity is created."""

    class AttributeChanged(
        Event[TTimestampedVersionedEntity],
        TimestampedEntity.AttributeChanged[TTimestampedVersionedEntity],
        VersionedEntity.AttributeChanged[TTimestampedVersionedEntity],
    ):
        """Published when a TimestampedVersionedEntity is created."""

    class Discarded(
        Event[TTimestampedVersionedEntity],
        TimestampedEntity.Discarded[TTimestampedVersionedEntity],
        VersionedEntity.Discarded[TTimestampedVersionedEntity],
    ):
        """Published when a TimestampedVersionedEntity is discarded."""


class TimeuuidedVersionedEntity(TimeuuidedEntity, VersionedEntity):
    pass



from abc import abstractmethod
from typing import Generic, Optional
from uuid import UUID

from eventsourcing.domain.model.events import AbstractSnapshot
from eventsourcing.whitehead import TEntity, TEvent


class AbstractEntityRepository(Generic[TEntity, TEvent]):
    @abstractmethod
    def __getitem__(self, entity_id: UUID) -> TEntity:
        """
        Returns entity for given ID.

        Raises ``RepositoryKeyError`` when entity not found.
        """

    @abstractmethod
    def __contains__(self, entity_id: UUID) -> bool:
        """
        Returns True or False, according to whether or not entity exists.
        """

    @abstractmethod
    def get_entity(
        self, entity_id: UUID, at: Optional[int] = None
    ) -> Optional[TEntity]:
        """
        Returns entity for given ID.

        Returns None when entity not found.
        """

    @abstractmethod
    def take_snapshot(
        self, entity_id: UUID, lt: Optional[int] = None, lte: Optional[int] = None
    ) -> Optional[AbstractSnapshot]:
        """
        Takes snapshot of entity state, using stored events.
        """



from typing import Any, Optional

from eventsourcing.infrastructure.base import AbstractRecordManager
from eventsourcing.infrastructure.django.manager import DjangoRecordManager
from eventsourcing.infrastructure.factory import InfrastructureFactory


class DjangoInfrastructureFactory(InfrastructureFactory):
    """
    Infrastructure factory for Django.
    """

    record_manager_class = DjangoRecordManager
    tracking_record_class: Any = None

    def __init__(
        self, tracking_record_class: Optional[type] = None, *args: Any, **kwargs: Any
    ):
        super(DjangoInfrastructureFactory, self).__init__(*args, **kwargs)
        self._tracking_record_class = tracking_record_class

    @property
    def integer_sequenced_record_class(self) -> Optional[type]:  # type: ignore
        from eventsourcing.infrastructure.django.models import IntegerSequencedRecord

        return IntegerSequencedRecord

    @property
    def timestamp_sequenced_record_class(self) -> Optional[type]:  # type: ignore
        from eventsourcing.infrastructure.django.models import TimestampSequencedRecord

        return TimestampSequencedRecord

    @property
    def snapshot_record_class(self) -> Optional[type]:  # type: ignore
        from eventsourcing.infrastructure.django.models import SnapshotRecord

        return SnapshotRecord

    def construct_integer_sequenced_record_manager(
        self, **kwargs: Any
    ) -> AbstractRecordManager:
        """
        Constructs Django record manager.

        :return: A Django record manager.
        :rtype: DjangoRecordManager
        """
        tracking_record_class = (
            self._tracking_record_class or self.tracking_record_class
        )
        return super(
            DjangoInfrastructureFactory, self
        ).construct_integer_sequenced_record_manager(
            tracking_record_class=tracking_record_class, **kwargs
        )



import unittest
from uuid import uuid4

from eventsourcing.application.policies import PersistencePolicy, SnapshottingPolicy
from eventsourcing.domain.model.entity import TimestampedEntity, VersionedEntity
from eventsourcing.domain.model.events import publish
from eventsourcing.domain.model.repository import AbstractEntityRepository
from eventsourcing.infrastructure.base import AbstractEventStore

try:
    from unittest import mock
except:
    import mock


class TestPersistencePolicy(unittest.TestCase):
    def setUp(self):
        self.event_store: AbstractEventStore = mock.Mock(spec=AbstractEventStore)
        self.persistence_policy = PersistencePolicy(
            event_store=self.event_store, persist_event_type=VersionedEntity.Event
        )

    def tearDown(self):
        self.persistence_policy.close()

    def test_published_events_are_appended_to_event_store(self):
        # Check the event store's append method has NOT been called.
        self.assertEqual(0, self.event_store.store_events.call_count)

        # Publish a versioned entity event.
        entity_id = uuid4()
        domain_event1 = VersionedEntity.Event(
            originator_id=entity_id, originator_version=0
        )
        publish([domain_event1])

        # Check the append method has been called once with the domain event.
        self.assertEqual(1, self.event_store.store_events.call_count)
        self.event_store.store_events.assert_called_once_with([domain_event1])

        # Publish a timestamped entity event (should be ignored).
        domain_event2 = TimestampedEntity.Event(originator_id=entity_id)
        publish([domain_event2])

        # Check the append() has still only been called once with the first domain event.
        self.assertEqual(1, self.event_store.store_events.call_count)
        self.event_store.store_events.assert_called_once_with([domain_event1])


class TestSnapshottingPolicy(unittest.TestCase):
    def setUp(self):
        self.repository: AbstractEntityRepository = mock.Mock(
            spec=AbstractEntityRepository
        )
        self.snapshot_store: AbstractEventStore = mock.Mock(spec=AbstractEventStore)
        self.policy = SnapshottingPolicy(
            repository=self.repository, snapshot_store=self.snapshot_store, period=2
        )

    def tearDown(self):
        self.policy.close()

    def test_published_events_are_appended_to_event_store(self):
        # Check the event store's append method has NOT been called.
        self.assertEqual(0, self.repository.take_snapshot.call_count)

        # Publish a versioned entity event.
        entity_id = uuid4()
        domain_event1 = VersionedEntity.Event(
            originator_id=entity_id, originator_version=0
        )
        domain_event2 = VersionedEntity.Event(
            originator_id=entity_id, originator_version=1
        )

        # Check take_snapshot is called once for each event.
        publish([domain_event1])
        self.assertEqual(0, self.repository.take_snapshot.call_count)
        publish([domain_event2])
        self.assertEqual(1, self.repository.take_snapshot.call_count)

        # Check take_snapshot is called once for each list.
        publish([domain_event1, domain_event2])
        self.assertEqual(2, self.repository.take_snapshot.call_count)



from json import JSONDecoder, JSONEncoder
from typing import Any, Generic, NamedTuple, Optional, Type

from eventsourcing.infrastructure.base import (
    DEFAULT_PIPELINE_ID,
    AbstractEventStore,
    AbstractRecordManager,
)
from eventsourcing.infrastructure.datastore import AbstractDatastore
from eventsourcing.infrastructure.eventstore import EventStore
from eventsourcing.infrastructure.sequenceditem import SequencedItem
from eventsourcing.infrastructure.sequenceditemmapper import (
    AbstractSequencedItemMapper,
    SequencedItemMapper,
)
from eventsourcing.utils.cipher.aes import AESCipher
from eventsourcing.whitehead import TEvent


class InfrastructureFactory(Generic[TEvent]):
    """
    Base class for infrastructure factories.
    """

    record_manager_class: Optional[Type[AbstractRecordManager]] = None
    sequenced_item_class: Type[NamedTuple] = SequencedItem  # type: ignore
    sequenced_item_mapper_class: Type[AbstractSequencedItemMapper] = SequencedItemMapper
    integer_sequenced_record_class: Optional[type] = None
    integer_sequenced_noid_record_class: Optional[type] = None
    timestamp_sequenced_record_class: Optional[type] = None
    snapshot_record_class: Optional[type] = None
    json_decoder_class: Optional[Type[JSONDecoder]] = None
    json_encoder_class: Optional[Type[JSONEncoder]] = None
    event_store_class: Optional[Type[AbstractEventStore]] = None

    def __init__(
        self,
        record_manager_class: Optional[Type[AbstractRecordManager]] = None,
        sequenced_item_class: Optional[Type[NamedTuple]] = None,
        event_store_class: Optional[Type[AbstractEventStore]] = None,
        sequenced_item_mapper_class: Optional[Type[AbstractSequencedItemMapper]] = None,
        json_encoder_class: Optional[Type[JSONEncoder]] = None,
        sort_keys: bool = False,
        json_decoder_class: Optional[Type[JSONDecoder]] = None,
        integer_sequenced_record_class: Optional[type] = None,
        timestamp_sequenced_record_class: Optional[type] = None,
        snapshot_record_class: Optional[type] = None,
        contiguous_record_ids: bool = False,
        application_name: Optional[str] = None,
        pipeline_id: int = DEFAULT_PIPELINE_ID,
    ):
        self.record_manager_class = (
            record_manager_class or type(self).record_manager_class
        )
        self.event_store_class = event_store_class or type(self).event_store_class
        self.sequenced_item_class = (
            sequenced_item_class or type(self).sequenced_item_class
        )
        self.sequenced_item_mapper_class = (
            sequenced_item_mapper_class or type(self).sequenced_item_mapper_class
        )
        self.json_encoder_class = json_encoder_class or type(self).json_encoder_class
        self.sort_keys = sort_keys
        self.json_decoder_class = json_decoder_class or type(self).json_decoder_class

        self._integer_sequenced_record_class = integer_sequenced_record_class

        self._timestamp_sequenced_record_class = timestamp_sequenced_record_class

        self._snapshot_record_class = snapshot_record_class

        self.contiguous_record_ids = contiguous_record_ids
        self.application_name = application_name
        self.pipeline_id = pipeline_id

    def construct_integer_sequenced_record_manager(
        self, **kwargs: Any
    ) -> AbstractRecordManager:
        """
        Constructs an integer sequenced record manager.
        """
        integer_sequenced_record_class = (
            self._integer_sequenced_record_class or self.integer_sequenced_record_class
        )
        return self.construct_record_manager(integer_sequenced_record_class, **kwargs)

    def construct_timestamp_sequenced_record_manager(
        self, **kwargs: Any
    ) -> AbstractRecordManager:
        """
        Constructs a timestamp sequenced record manager.
        """
        timestamp_sequenced_record_class = (
            self._timestamp_sequenced_record_class
            or self.timestamp_sequenced_record_class
        )
        return self.construct_record_manager(timestamp_sequenced_record_class, **kwargs)

    def construct_snapshot_record_manager(self, **kwargs: Any) -> AbstractRecordManager:
        """
        Constructs a snapshot record manager.
        """
        snapshot_record_class = (
            self._snapshot_record_class or self.snapshot_record_class
        )
        return self.construct_record_manager(snapshot_record_class, **kwargs)

    def construct_record_manager(
        self,
        record_class: Optional[type],
        sequenced_item_class: Optional[Type[NamedTuple]] = None,
        **kwargs: Any
    ) -> AbstractRecordManager:
        """
        Constructs an record manager.
        """
        assert self.record_manager_class is not None
        return self.record_manager_class(
            sequenced_item_class=sequenced_item_class or self.sequenced_item_class,
            record_class=record_class,
            contiguous_record_ids=self.contiguous_record_ids,
            application_name=self.application_name,
            pipeline_id=self.pipeline_id,
            **kwargs
        )

    def construct_sequenced_item_mapper(
        self, cipher: Optional[AESCipher], compressor: Any,
    ) -> AbstractSequencedItemMapper:
        """
        Constructs sequenced item mapper object.

        :returns: Sequenced item mapper object.
        :rtype: eventsourcing.infrastructure.sequenceditemmapper
        .AbstractSequencedItemMapper
        """
        return self.sequenced_item_mapper_class(
            sequenced_item_class=self.sequenced_item_class,
            cipher=cipher,
            compressor=compressor,
            # sequence_id_attr_name=sequence_id_attr_name,
            # position_attr_name=position_attr_name,
            json_encoder_class=self.json_encoder_class,
            sort_keys=self.sort_keys,
            json_decoder_class=self.json_decoder_class,
        )

    def construct_integer_sequenced_event_store(
        self, cipher: Optional[AESCipher], compressor: Any,
    ) -> AbstractEventStore:
        """
        Constructs an integer sequenced event store.
        """
        sequenced_item_mapper = self.construct_sequenced_item_mapper(cipher, compressor)
        record_manager = self.construct_integer_sequenced_record_manager()
        return (self.event_store_class or EventStore)(
            record_manager=record_manager, event_mapper=sequenced_item_mapper
        )

    def construct_datastore(self) -> Optional[AbstractDatastore]:
        """
        Constructs datastore object.

        :returns: Concrete datastore object object.
        """
        return None



import django
from django.db import models
from django.db.backends.base.schema import BaseDatabaseSchemaEditor

from eventsourcing.application.django import DjangoApplication
from eventsourcing.tests.sequenced_item_tests.test_django_record_manager import (
    DjangoTestCase,
)
from eventsourcing.tests.test_process import TestProcessApplication


class TestProcessWithDjango(DjangoTestCase, TestProcessApplication):
    infrastructure_class = DjangoApplication

    def test_projection_into_custom_orm_obj(self):
        super(TestProcessWithDjango, self).test_projection_into_custom_orm_obj()

    def define_projection_record_class(self):
        class ProjectionRecord(models.Model):
            uid = models.BigAutoField(primary_key=True)

            # Sequence ID (e.g. an entity or aggregate ID).
            projection_id = models.UUIDField()

            # State of the item (serialized dict, possibly encrypted).
            state = models.TextField()

            class Meta:
                db_table = "projections"
                app_label = "projections"
                managed = False

        self.projection_record_class = ProjectionRecord

    def setup_projections_table(self, process):
        from django.db import connections

        with connections["default"].schema_editor() as schema_editor:
            assert isinstance(schema_editor, BaseDatabaseSchemaEditor)
            try:
                schema_editor.delete_model(self.projection_record_class)
            except django.db.utils.ProgrammingError:
                pass

        with connections["default"].schema_editor() as schema_editor:
            assert isinstance(schema_editor, BaseDatabaseSchemaEditor)
            schema_editor.create_model(self.projection_record_class)

    def get_projection_record(
        self, projection_record_class, projection_id, record_manager
    ):
        try:
            return projection_record_class.objects.get(projection_id=projection_id)
        except projection_record_class.DoesNotExist:
            return None


del TestProcessApplication



import datetime
from decimal import Decimal
from typing import Any, Optional, Union
from uuid import UUID, uuid5

from dateutil.relativedelta import relativedelta

from eventsourcing.domain.model.entity import TimestampedVersionedEntity
from eventsourcing.domain.model.events import (
    EventWithOriginatorID,
    EventWithTimestamp,
    LoggedEvent,
    publish,
)
from eventsourcing.domain.model.repository import AbstractEntityRepository
from eventsourcing.exceptions import RepositoryKeyError
from eventsourcing.utils.times import (
    datetime_from_timestamp,
    decimaltimestamp,
    utc_timezone,
)
from eventsourcing.utils.topic import get_topic

Namespace_Timebuckets = UUID("0d7ee297-a976-4c29-91ff-84ffc79d8155")

ONE_YEAR = relativedelta(years=1)
ONE_MONTH = relativedelta(months=1)
ONE_DAY = relativedelta(days=1)
ONE_HOUR = relativedelta(hours=1)
ONE_MINUTE = relativedelta(minutes=1)
ONE_SECOND = relativedelta(seconds=1)

BUCKET_SIZES = {
    "year": ONE_YEAR,
    "month": ONE_MONTH,
    "day": ONE_DAY,
    "hour": ONE_HOUR,
    "minute": ONE_MINUTE,
    "second": ONE_SECOND,
}


class Timebucketedlog(TimestampedVersionedEntity):
    class Event(TimestampedVersionedEntity.Event):
        """Supertype for events of time-bucketed log."""

    class Started(TimestampedVersionedEntity.Created, Event):
        pass

    class BucketSizeChanged(Event, TimestampedVersionedEntity.AttributeChanged):
        pass

    def __init__(self, name: UUID, bucket_size: Optional[str] = None, **kwargs: Any):
        super(Timebucketedlog, self).__init__(**kwargs)
        self._name = name
        self._bucket_size = bucket_size

    @property
    def name(self) -> UUID:
        return self._name

    @property
    def started_on(self) -> Decimal:
        return self.__created_on__

    @property
    def bucket_size(self) -> str:
        assert self._bucket_size
        return self._bucket_size

    def log_message(self, message: str) -> "MessageLogged":
        assert isinstance(message, str)
        bucket_id = make_timebucket_id(self.name, decimaltimestamp(), self.bucket_size)
        event = MessageLogged(originator_id=bucket_id, message=message)
        publish([event])
        return event


class TimebucketedlogRepository(AbstractEntityRepository):
    def get_or_create(self, log_name: UUID, bucket_size: str) -> Timebucketedlog:
        """
        Gets or creates a log.
        """
        try:
            return self[log_name]
        except RepositoryKeyError:
            return start_new_timebucketedlog(log_name, bucket_size=bucket_size)


def start_new_timebucketedlog(
    name: UUID, bucket_size: Optional[str] = None
) -> Timebucketedlog:
    if bucket_size is None:
        bucket_size = "year"
    if bucket_size not in BUCKET_SIZES:
        raise ValueError(
            "Bucket size '{}' not supported, must be one of: {}"
            "".format(bucket_size, BUCKET_SIZES.keys())
        )
    event = Timebucketedlog.Started(
        originator_id=name,
        name=name,
        bucket_size=bucket_size,
        originator_topic=get_topic(Timebucketedlog),
    )
    entity = event.__mutate__(None)
    assert entity is not None
    publish([event])
    return entity


class MessageLogged(EventWithTimestamp, EventWithOriginatorID, LoggedEvent):
    def __init__(self, message: str, originator_id: UUID):
        super(MessageLogged, self).__init__(
            originator_id=originator_id, message=message
        )

    @property
    def message(self) -> str:
        return self.__dict__["message"]


def make_timebucket_id(
    log_id: UUID, timestamp: Union[Decimal, float], bucket_size: str
) -> UUID:
    d = datetime_from_timestamp(timestamp)

    assert isinstance(d, datetime.datetime)
    if bucket_size.startswith("year"):
        boundary = "{:04}".format(d.year)
    elif bucket_size.startswith("month"):
        boundary = "{:04}-{:02}".format(d.year, d.month)
    elif bucket_size.startswith("day"):
        boundary = "{:04}-{:02}-{:02}".format(d.year, d.month, d.day)
    elif bucket_size.startswith("hour"):
        boundary = "{:04}-{:02}-{:02}_{:02}".format(d.year, d.month, d.day, d.hour)
    elif bucket_size.startswith("minute"):
        boundary = "{:04}-{:02}-{:02}_{:02}-{:02}".format(
            d.year, d.month, d.day, d.hour, d.minute
        )
    elif bucket_size.startswith("second"):
        boundary = "{:04}-{:02}-{:02}_{:02}-{:02}-{:02}".format(
            d.year, d.month, d.day, d.hour, d.minute, d.second
        )
    else:
        raise ValueError("Bucket size not supported: {}".format(bucket_size))
    return uuid5(Namespace_Timebuckets, log_id.hex + "_" + boundary)


def next_bucket_starts(timestamp: float, bucket_size: str) -> float:
    starts = bucket_starts(timestamp, bucket_size)
    duration = bucket_duration(bucket_size)
    return timestamp_from_datetime((starts + duration))


def previous_bucket_starts(timestamp: float, bucket_size: str) -> float:
    starts = bucket_starts(timestamp, bucket_size)
    duration = bucket_duration(bucket_size)
    return timestamp_from_datetime((starts - duration))


def bucket_starts(timestamp: float, bucket_size: str) -> datetime.datetime:
    dt = datetime_from_timestamp(timestamp)
    assert isinstance(dt, datetime.datetime)
    if bucket_size.startswith("year"):
        return datetime.datetime(dt.year, 1, 1, tzinfo=utc_timezone)
    elif bucket_size.startswith("month"):
        return datetime.datetime(dt.year, dt.month, 1, tzinfo=utc_timezone)
    elif bucket_size.startswith("day"):
        return datetime.datetime(dt.year, dt.month, dt.day, tzinfo=utc_timezone)
    elif bucket_size.startswith("hour"):
        return datetime.datetime(
            dt.year, dt.month, dt.day, dt.hour, tzinfo=utc_timezone
        )
    elif bucket_size.startswith("minute"):
        return datetime.datetime(
            dt.year, dt.month, dt.day, dt.hour, dt.minute, tzinfo=utc_timezone
        )
    elif bucket_size.startswith("second"):
        return datetime.datetime(
            dt.year,
            dt.month,
            dt.day,
            dt.hour,
            dt.minute,
            dt.second,
            tzinfo=utc_timezone,
        )
    else:
        raise ValueError("Bucket size not supported: {}".format(bucket_size))


def bucket_duration(bucket_size: str) -> relativedelta:
    try:
        return BUCKET_SIZES[bucket_size]
    except KeyError:
        raise ValueError(
            "Bucket size not supported: {}. Must be one of: {}"
            "".format(bucket_size, BUCKET_SIZES.keys())
        )


# Todo: Move to general utils?
def timestamp_from_datetime(dt: datetime.datetime) -> float:
    assert dt.tzinfo, "Datetime object does not have tzinfo"
    return dt.timestamp()



from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="EntitySnapshotRecord",
            fields=[
                ("uid", models.BigAutoField(primary_key=True, serialize=False)),
                ("application_name", models.CharField(max_length=32)),
                ("originator_id", models.UUIDField()),
                ("originator_version", models.BigIntegerField()),
                ("topic", models.TextField()),
                ("state", models.BinaryField()),
            ],
            options={"db_table": "entity_snapshots"},
        ),
        migrations.CreateModel(
            name="IntegerSequencedRecord",
            fields=[
                ("id", models.BigAutoField(primary_key=True, serialize=False)),
                ("sequence_id", models.UUIDField()),
                ("position", models.BigIntegerField()),
                ("topic", models.TextField()),
                ("state", models.BinaryField()),
            ],
            options={"db_table": "integer_sequenced_items"},
        ),
        migrations.CreateModel(
            name="NotificationTrackingRecord",
            fields=[
                ("uid", models.BigAutoField(primary_key=True, serialize=False)),
                ("application_name", models.CharField(max_length=32)),
                ("upstream_application_name", models.CharField(max_length=32)),
                ("pipeline_id", models.IntegerField()),
                ("notification_id", models.BigIntegerField()),
            ],
            options={"db_table": "notification_tracking"},
        ),
        migrations.CreateModel(
            name="SnapshotRecord",
            fields=[
                ("uid", models.BigAutoField(primary_key=True, serialize=False)),
                ("sequence_id", models.UUIDField()),
                ("position", models.BigIntegerField()),
                ("topic", models.TextField()),
                ("state", models.BinaryField()),
            ],
            options={"db_table": "snapshots"},
        ),
        migrations.CreateModel(
            name="StoredEventRecord",
            fields=[
                ("uid", models.BigAutoField(primary_key=True, serialize=False)),
                ("application_name", models.CharField(max_length=32)),
                ("originator_id", models.UUIDField()),
                ("originator_version", models.BigIntegerField()),
                ("pipeline_id", models.IntegerField()),
                ("notification_id", models.BigIntegerField()),
                ("topic", models.TextField()),
                ("state", models.BinaryField()),
                ("causal_dependencies", models.TextField()),
            ],
            options={"db_table": "stored_events"},
        ),
        migrations.CreateModel(
            name="TimestampSequencedRecord",
            fields=[
                ("id", models.BigAutoField(primary_key=True, serialize=False)),
                ("sequence_id", models.UUIDField()),
                ("position", models.DecimalField(decimal_places=6, max_digits=24)),
                ("topic", models.TextField()),
                ("state", models.BinaryField()),
            ],
            options={"db_table": "timestamp_sequenced_items"},
        ),
        migrations.AddIndex(
            model_name="timestampsequencedrecord",
            index=models.Index(fields=["position"], name="position_idx"),
        ),
        migrations.AlterUniqueTogether(
            name="timestampsequencedrecord",
            unique_together={("sequence_id", "position")},
        ),
        migrations.AlterUniqueTogether(
            name="storedeventrecord",
            unique_together={
                ("application_name", "originator_id", "originator_version"),
                ("application_name", "pipeline_id", "notification_id"),
            },
        ),
        migrations.AlterUniqueTogether(
            name="snapshotrecord", unique_together={("sequence_id", "position")}
        ),
        migrations.AlterUniqueTogether(
            name="notificationtrackingrecord",
            unique_together={
                (
                    "application_name",
                    "upstream_application_name",
                    "pipeline_id",
                    "notification_id",
                )
            },
        ),
        migrations.AlterUniqueTogether(
            name="integersequencedrecord", unique_together={("sequence_id", "position")}
        ),
        migrations.AlterUniqueTogether(
            name="entitysnapshotrecord",
            unique_together={
                ("application_name", "originator_id", "originator_version")
            },
        ),
    ]



from typing import Any, Dict, Iterable, Optional, Sequence
from uuid import UUID

from django.db import IntegrityError, ProgrammingError, connection, transaction

from eventsourcing.infrastructure.base import SQLRecordManager, TrackingKwargs


class DjangoRecordManager(SQLRecordManager):
    _where_application_name_tmpl = " WHERE application_name = %s AND pipeline_id = %s"

    def write_records(
        self,
        records: Iterable[Any],
        tracking_kwargs: Optional[TrackingKwargs] = None,
        orm_objs_pending_save: Optional[Sequence[Any]] = None,
        orm_objs_pending_delete: Optional[Sequence[Any]] = None,
    ) -> None:
        try:
            with transaction.atomic(self.record_class.objects.db):  # type: ignore
                with connection.cursor() as cursor:
                    # Insert tracking record.
                    if tracking_kwargs:
                        params = [
                            tracking_kwargs[c] for c in self.tracking_record_field_names
                        ]
                        cursor.execute(self.insert_tracking_record, params)

                    if self.contiguous_record_ids:
                        # Use cursor to execute insert select max statement.
                        for record in records:
                            # Get values from record obj.
                            # List of params, because dict doesn't work with Django
                            # and SQLite.
                            params = []

                            for col_name in self.field_names:
                                col_value = getattr(record, col_name)
                                meta = self.record_class._meta  # type: ignore
                                col_type = meta.get_field(col_name)

                                # Prepare value for database.
                                param = col_type.get_db_prep_value(
                                    col_value, connection
                                )
                                params.append(param)

                            # Notification logs fields, to be inserted with event
                            # fields.
                            if hasattr(self.record_class, "application_name"):
                                params.append(self.application_name)
                            if hasattr(self.record_class, "pipeline_id"):
                                params.append(self.pipeline_id)
                            if hasattr(record, "causal_dependencies"):
                                params.append(record.causal_dependencies)

                            # Where clause fields.
                            if hasattr(self.record_class, "application_name"):
                                params.append(self.application_name)
                            if hasattr(self.record_class, "pipeline_id"):
                                params.append(self.pipeline_id)

                            # Execute insert statement.
                            cursor.execute(self.insert_select_max, params)
                            # Todo: Use insert_values when records have IDs (like
                            #  SQLAlchemy manager).
                            # Todo: Support 'event-not-notifiable' by setting
                            #  pipeline ID and notification ID to None.

                    else:
                        # This can only work for simple models, without application_name
                        # and pipeline_id, because it relies on the auto-incrementing
                        # ID.
                        # Todo: If it's faster, change to use an "insert_values" raw
                        #  query.
                        # Save record objects.
                        for record in records:
                            record.save()

                    # Call 'save()' on each of the ORM objects pending save.
                    if orm_objs_pending_save:
                        for orm_obj in orm_objs_pending_save:
                            orm_obj.save()

                    # Call 'delete()' on each of the ORM objects pending delete.
                    if orm_objs_pending_delete:
                        for orm_obj in orm_objs_pending_delete:
                            orm_obj.delete()

        except IntegrityError as e:
            self.raise_record_integrity_error(e)

    def make_placeholder(self, _: str) -> str:
        return "%s"  # doesn't involve field name

    def get_record_table_name(self, record_class: type) -> str:
        """Returns table name from SQLAlchemy record class."""
        return record_class._meta.db_table  # type: ignore

    def get_record(self, sequence_id: UUID, position: int) -> Any:
        kwargs = {
            self.field_names.sequence_id: sequence_id,
            self.field_names.position: position,
        }
        records = self.record_class.objects.filter(**kwargs)  # type: ignore
        # Todo: try/except for native error here, call self.raise_index_error()
        return records.all()[0]

    def get_records(
        self,
        sequence_id: UUID,
        gt: Optional[int] = None,
        gte: Optional[int] = None,
        lt: Optional[int] = None,
        lte: Optional[int] = None,
        limit: Optional[int] = None,
        query_ascending: bool = True,
        results_ascending: bool = True,
    ) -> Sequence[Any]:

        assert limit is None or limit >= 1, limit

        filter_kwargs = {self.field_names.sequence_id: sequence_id}
        objects = self.record_class.objects.filter(**filter_kwargs)  # type: ignore

        if hasattr(self.record_class, "application_name"):
            objects = objects.filter(application_name=self.application_name)

        position_field_name = self.field_names.position

        if query_ascending:
            objects = objects.order_by(position_field_name)
        else:
            objects = objects.order_by("-" + position_field_name)

        if gt is not None:
            arg = "{}__gt".format(position_field_name)
            objects = objects.filter(**{arg: gt})
        if gte is not None:
            arg = "{}__gte".format(position_field_name)
            objects = objects.filter(**{arg: gte})
        if lt is not None:
            arg = "{}__lt".format(position_field_name)
            objects = objects.filter(**{arg: lt})
        if lte is not None:
            arg = "{}__lte".format(position_field_name)
            objects = objects.filter(**{arg: lte})

        if limit is not None:
            objects = objects[:limit]

        records = objects.all()

        if results_ascending != query_ascending:
            # This code path is under test, but not otherwise used ATM.
            records = list(records)
            records.reverse()

        for record in records:
            # Django returns memoryview objects from PostgreSQL, so need to cast.
            state = getattr(record, self.field_names.state)
            setattr(record, self.field_names.state, bytes(state))

        return records

    def get_notification_records(
        self,
        start: Optional[int] = None,
        stop: Optional[int] = None,
        *args: Any,
        **kwargs: Any
    ) -> Iterable:
        """
        Returns all records in the table.
        """
        filter_kwargs = {}
        # Todo: Also support sequencing by 'position' if items are sequenced by
        #  timestamp?
        if start is not None:
            filter_kwargs["%s__gte" % self.notification_id_name] = start + 1
        if stop is not None:
            filter_kwargs["%s__lt" % self.notification_id_name] = stop + 1
        objects = self.record_class.objects.filter(**filter_kwargs)  # type: ignore

        if hasattr(self.record_class, "application_name"):
            objects = objects.filter(application_name=self.application_name)
        if hasattr(self.record_class, "pipeline_id"):
            objects = objects.filter(pipeline_id=self.pipeline_id)

        objects = objects.order_by("%s" % self.notification_id_name)
        for record in objects.all():
            # Django returns memoryview objects from PostgreSQL, so need to cast.
            state = getattr(record, self.field_names.state)
            setattr(record, self.field_names.state, bytes(state))
            yield record

    def delete_record(self, record: Any) -> None:
        """
        Permanently removes record from table.
        """
        record.delete()

    def get_max_notification_id(self) -> int:
        assert self.notification_id_name
        try:
            objects = self.record_class.objects  # type: ignore
            if hasattr(self.record_class, "application_name"):
                objects = objects.filter(application_name=self.application_name)
            if hasattr(self.record_class, "pipeline_id"):
                objects = objects.filter(pipeline_id=self.pipeline_id)
            latest = objects.latest(self.notification_id_name)
            return getattr(latest, self.notification_id_name)
        except (self.record_class.DoesNotExist, ProgrammingError):  # type: ignore
            return 0

    def get_max_tracking_record_id(self, upstream_application_name: str) -> int:
        notification_id = 0
        assert self.tracking_record_class is not None
        try:
            objects = self.tracking_record_class.objects  # type: ignore
            objects = objects.filter(application_name=self.application_name)
            objects = objects.filter(
                upstream_application_name=upstream_application_name
            )
            objects = objects.filter(pipeline_id=self.pipeline_id)
            notification_id = objects.latest("notification_id").notification_id
        except self.tracking_record_class.DoesNotExist:  # type: ignore
            pass
        return notification_id

    def has_tracking_record(
        self, upstream_application_name: str, pipeline_id: int, notification_id: int
    ) -> bool:
        objects = self.tracking_record_class.objects  # type: ignore
        objects = objects.filter(application_name=self.application_name)
        objects = objects.filter(upstream_application_name=upstream_application_name)
        objects = objects.filter(pipeline_id=pipeline_id)
        objects = objects.filter(notification_id=notification_id)
        return bool(objects.count())

    def all_sequence_ids(self) -> Iterable[UUID]:
        sequence_id_fieldname = self.field_names.sequence_id
        values_queryset = self.record_class.objects.values(  # type: ignore
            sequence_id_fieldname
        ).distinct()
        for values in values_queryset:
            yield values[sequence_id_fieldname]

    def get_field_kwargs(self, item: object) -> Dict[str, Any]:
        kwargs = super().get_field_kwargs(item)
        state_fieldname = self.field_names.state
        state = kwargs[state_fieldname]
        if isinstance(state, memoryview):
            state = bytes(state)
            kwargs[state_fieldname] = state
        return kwargs



from uuid import UUID

from eventsourcing.application.policies import PersistencePolicy
from eventsourcing.domain.model.events import DomainEvent
from eventsourcing.example.domainmodel import create_new_example
from eventsourcing.example.infrastructure import ExampleRepository
from eventsourcing.infrastructure.cassandra.datastore import (
    CassandraDatastore,
    CassandraSettings,
)
from eventsourcing.infrastructure.cassandra.manager import CassandraRecordManager
from eventsourcing.infrastructure.cassandra.records import StoredEventRecord
from eventsourcing.infrastructure.eventstore import EventStore
from eventsourcing.infrastructure.sequenceditem import StoredEvent
from eventsourcing.infrastructure.sequenceditemmapper import SequencedItemMapper
from eventsourcing.tests.datastore_tests.base import AbstractDatastoreTestCase

# In this test the default SequencedItem class is replaced with a "stored event" class.
# How easy is it to customize the infrastructure to support that? We just need
# to define the new sequenced item class, define a suitable record class,
# and configure the other components. It's easy.


class ExampleApplicationWithAlternativeSequencedItemType(object):
    def __init__(self):
        self.event_store = EventStore(
            record_manager=CassandraRecordManager(
                record_class=StoredEventRecord, sequenced_item_class=StoredEvent
            ),
            event_mapper=SequencedItemMapper(
                sequenced_item_class=StoredEvent, other_attr_names=()
            ),
        )
        self.repository = ExampleRepository(event_store=self.event_store)
        self.persistence_policy = PersistencePolicy(
            event_store=self.event_store, persist_event_type=DomainEvent
        )

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.persistence_policy.close()


class TestExampleWithAlternativeSequencedItemType(AbstractDatastoreTestCase):
    def setUp(self):
        super(TestExampleWithAlternativeSequencedItemType, self).setUp()
        self.datastore.setup_connection()
        self.datastore.setup_tables()

    def tearDown(self):
        self.datastore.drop_tables()
        self.datastore.close_connection()
        super(TestExampleWithAlternativeSequencedItemType, self).setUp()

    def construct_datastore(self):
        return CassandraDatastore(
            settings=CassandraSettings(), tables=(StoredEventRecord,)
        )

    def test(self):
        with ExampleApplicationWithAlternativeSequencedItemType() as app:
            # Create entity.
            entity1 = create_new_example(a="a", b="b")
            self.assertIsInstance(entity1.id, UUID)
            self.assertEqual(entity1.a, "a")
            self.assertEqual(entity1.b, "b")

            # Check there is a stored event.
            all_records = list(app.event_store.record_manager.get_records(entity1.id))
            assert len(all_records) == 1, len(all_records)
            stored_event = all_records[0]
            assert isinstance(stored_event, StoredEventRecord), stored_event
            assert stored_event.originator_id == entity1.id
            assert stored_event.originator_version == 0

            # Read entity from repo.
            retrieved_obj = app.repository[entity1.id]
            self.assertEqual(retrieved_obj.id, entity1.id)



# from eventsourcing.contrib.suffixtrees.domain.model.generalizedsuffixtree import (
#     EdgeRepository,
#     GeneralizedSuffixTree,
#     GeneralizedSuffixTreeRepository,
#     NodeRepository,
#     StringidCollection,
#     SuffixTreeEdge,
#     SuffixTreeNode,
#     SuffixTreeNodeChildCollection,
# )
# from eventsourcing.infrastructure.repositories.collection_repo import (
#     CollectionRepository,
# )
# from eventsourcing.infrastructure.eventsourcedrepository import EventSourcedRepository
#
#
# class GeneralizedSuffixTreeRepo(
#     EventSourcedRepository, GeneralizedSuffixTreeRepository
# ):
#     domain_class = GeneralizedSuffixTree
#
#
# class NodeRepo(EventSourcedRepository, NodeRepository):
#     domain_class = SuffixTreeNode
#
#
# class NodeChildCollectionRepo(EventSourcedRepository, NodeRepository):
#     domain_class = SuffixTreeNodeChildCollection
#
#
# class EdgeRepo(EventSourcedRepository, EdgeRepository):
#     domain_class = SuffixTreeEdge
#
#
# class StringidCollectionRepo(CollectionRepository):
#     domain_class = StringidCollection



import os
from typing import Optional
from uuid import uuid4

from redis import Redis, StrictRedis

from eventsourcing.infrastructure.integersequencegenerators.base import (
    AbstractIntegerSequenceGenerator,
)


class RedisIncr(AbstractIntegerSequenceGenerator):
    """
    Generates a sequence of integers, using Redis' INCR command.

    Maximum number is 2**63, or 9223372036854775807, the maximum
    value of a 64 bit signed integer.
    """

    def __init__(self, redis: Optional[Redis] = None, key: Optional[str] = None):
        redis_host = os.getenv("REDIS_HOST", "localhost")
        self.redis = redis or StrictRedis(host=redis_host)
        self.key = key or "integer-sequence-generator-{}".format(uuid4())

    def __next__(self) -> int:
        return self.redis.incr(self.key) - 1



from eventsourcing.application.django import DjangoApplication
from eventsourcing.tests.sequenced_item_tests.test_django_record_manager import (
    DjangoTestCase,
)
from eventsourcing.tests.test_thespian_runner import TestThespianRunner


class TestThespianRunnerWithDjango(DjangoTestCase, TestThespianRunner):
    infrastructure_class = DjangoApplication



from typing import Any, Iterator, Optional, Set
from uuid import UUID

from eventsourcing.domain.model.entity import (
    TimestampedVersionedEntity,
    TTimestampedVersionedEntity,
)
from eventsourcing.domain.model.repository import AbstractEntityRepository


class Collection(TimestampedVersionedEntity):
    class Event(TimestampedVersionedEntity.Event[TTimestampedVersionedEntity]):
        """Supertype for events of collection entities."""

    class Created(
        Event[TTimestampedVersionedEntity],
        TimestampedVersionedEntity.Created[TTimestampedVersionedEntity],
    ):
        """Published when collection is created."""

    class Discarded(
        Event[TTimestampedVersionedEntity],
        TimestampedVersionedEntity.Discarded[TTimestampedVersionedEntity],
    ):
        """Published when collection is discarded."""

    class EventWithItem(Event[TTimestampedVersionedEntity]):
        @property
        def item(self) -> Any:
            return self.__dict__["item"]

    def __init__(self, **kwargs: Any):
        super(Collection, self).__init__(**kwargs)
        self._items: Set = set()

    def __iter__(self) -> Iterator:
        return self._items.__iter__()

    @property
    def items(self) -> Set:
        self.__assert_not_discarded__()
        return self._items

    def add_item(self, item: Any) -> None:
        self.__trigger_event__(self.ItemAdded, item=item)

    class ItemAdded(EventWithItem):
        def mutate(self, obj: "Collection") -> None:
            obj._items.add(self.item)

    def remove_item(self, item: Any) -> None:
        self.__trigger_event__(self.ItemRemoved, item=item)

    class ItemRemoved(EventWithItem):
        def mutate(self, obj: "Collection") -> None:
            obj._items.remove(self.item)


def register_new_collection(collection_id: Optional[UUID] = None) -> Collection:
    return Collection.__create__(originator_id=collection_id)


class AbstractCollectionRepository(AbstractEntityRepository):
    pass



import datetime
from collections import ChainMap, OrderedDict, deque, namedtuple
from decimal import Decimal
from enum import Enum
from fractions import Fraction
from unittest import TestCase, skipIf
from uuid import NAMESPACE_URL, UUID

from eventsourcing.utils.times import utc_timezone
from eventsourcing.utils.transcoding import ObjectJSONDecoder, ObjectJSONEncoder

try:
    from dataclasses import make_dataclass
except:
    make_dataclass = None


class TestTranscoding(TestCase):
    def test_str(self):
        value = "a string"
        encoded = '"a string"'
        self.assertTranscoding(value, encoded)

    def test_str_type(self):
        value = str
        encoded = '{"__type__":"builtins#str"}'
        self.assertTranscoding(value, encoded)

    def test_int(self):
        value = 1
        encoded = "1"
        self.assertTranscoding(value, encoded)

    def test_int_type(self):
        value = int
        encoded = '{"__type__":"builtins#int"}'
        self.assertTranscoding(value, encoded)

    def test_float(self):
        value = 1.001
        encoded = "1.001"
        self.assertTranscoding(value, encoded)

    def test_float_type(self):
        value = float
        encoded = '{"__type__":"builtins#float"}'
        self.assertTranscoding(value, encoded)

    def test_datetime_without_tzinfo(self):
        value = datetime.datetime(2011, 1, 1, 1, 1, 1)
        encoded = '{"ISO8601_datetime":"2011-01-01T01:01:01.000000"}'
        self.assertTranscoding(value, encoded)

    def test_datetime_with_tzinfo(self):
        value = datetime.datetime(2011, 1, 1, 1, 1, 1, tzinfo=utc_timezone)
        encoded = '{"ISO8601_datetime":"2011-01-01T01:01:01.000000+0000"}'
        self.assertTranscoding(value, encoded)

    def test_datetime_type(self):
        value = datetime.datetime
        encoded = '{"__type__":"datetime#datetime"}'
        self.assertTranscoding(value, encoded)

    def test_date(self):
        value = datetime.date(2011, 1, 1)
        encoded = '{"ISO8601_date":"2011-01-01"}'
        self.assertTranscoding(value, encoded)

    def test_date_type(self):
        value = datetime.date
        encoded = '{"__type__":"datetime#date"}'
        self.assertTranscoding(value, encoded)

    def test_time(self):
        value = datetime.time(23, 59, 59, 123456)
        encoded = '{"ISO8601_time":"23:59:59.123456"}'
        self.assertTranscoding(value, encoded)

    def test_time_type(self):
        value = datetime.time
        encoded = '{"__type__":"datetime#time"}'
        self.assertTranscoding(value, encoded)

    def test_decimal(self):
        value = Decimal("59.123456")
        encoded = '{"__decimal__":"59.123456"}'
        self.assertTranscoding(value, encoded)

    def test_decimal_type(self):
        value = Decimal
        encoded = '{"__type__":"decimal#Decimal"}'
        self.assertTranscoding(value, encoded)

    def test_fraction(self):
        value = Fraction(1, 3)
        encoded = (
            '{"__class__":{"state":{"_denominator":3,"_numerator":1},'
            '"topic":"fractions#Fraction"}}'
        )
        self.assertTranscoding(value, encoded)

    def test_fraction_type(self):
        value = Fraction
        encoded = '{"__type__":"fractions#Fraction"}'
        self.assertTranscoding(value, encoded)

    def test_enum(self):
        value = Colour.GREEN
        encoded = (
            '{"__enum__":{"name":"GREEN","topic":'
            '"eventsourcing.tests.test_transcoding#Colour"}}'
        )
        self.assertTranscoding(value, encoded)

    def test_uuid(self):
        value = NAMESPACE_URL
        encoded = '{"UUID":"6ba7b8119dad11d180b400c04fd430c8"}'
        self.assertTranscoding(value, encoded)

    def test_uuid_type(self):
        value = UUID
        encoded = '{"__type__":"uuid#UUID"}'
        self.assertTranscoding(value, encoded)

    def test_tuple(self):
        value = (1, 2, 4)
        encoded = '{"__tuple__":[1,2,4]}'
        self.assertTranscoding(value, encoded)

    def test_tuple_type(self):
        value = tuple
        encoded = '{"__type__":"builtins#tuple"}'
        self.assertTranscoding(value, encoded)

    def test_namedtuple(self):
        value = MyNamedTuple(a=1, b="2", c=MyObjectClass([3, "4"]))
        encoded = (
            '{"__tuple__":{"state":[1,"2",{"__class__":{"state":'
            '{"a":[3,"4"]},"topic":"eventsourcing.tests.test_trans'
            'coding#MyObjectClass"}}],"topic":"eventsourcing.tests.t'
            'est_transcoding#MyNamedTuple"}}'
        )
        self.assertTranscoding(value, encoded)

    def test_namedtuple_type(self):
        value = MyNamedTuple
        encoded = '{"__type__":"eventsourcing.tests.test_transcoding#MyNamedTuple"}'
        self.assertTranscoding(value, encoded)

    def test_tuple_of_tuples(self):
        value = ((1, 1), (2, 2))
        encoded = '{"__tuple__":[{"__tuple__":[1,1]},{"__tuple__":[2,2]}]}'
        self.assertTranscoding(value, encoded)

    def test_dict_with_tuple_values(self):
        value = {"1": (2, 3), "4": (5, 6)}
        encoded = '{"1":{"__tuple__":[2,3]},' '"4":{"__tuple__":[5,6]}}'
        self.assertTranscoding(value, encoded)

    def test_custom_dict_with_tuple_values(self):
        value = MyDict((("1", (2, 3)), ("4", (5, 6))))

        encoded = (
            '{"__dict__":{"state":{'
            '"1":{"__tuple__":[2,3]},'
            '"4":{"__tuple__":[5,6]}},'
            '"topic":"eventsourcing.tests.test_transcoding#MyDict"}}'
        )
        self.assertTranscoding(value, encoded)

    def test_dict_with_decimal_values(self):
        value = {"1": Decimal("1.2"), "4": Decimal("3.4")}
        encoded = '{"1":{"__decimal__":"1.2"},"4":{"__decimal__":"3.4"}}'
        self.assertTranscoding(value, encoded)

    def test_ordered_dict(self):
        value = OrderedDict(a=1, b=2, c=3)
        encoded = (
            '{"__dict__":{"state":{"a":1,"b":2,"c":3},'
            '"topic":"collections#OrderedDict"}}'
        )
        self.assertTranscoding(value, encoded)

    def test_ordered_dict_with_tuple(self):
        value = OrderedDict(a=(1, 1))
        encoded = (
            '{"__dict__":{'
            '"state":{"a":{"__tuple__":[1,1]}},'
            '"topic":"collections#OrderedDict"}}'
        )
        self.assertTranscoding(value, encoded)

    def test_chainmap(self):
        value = ChainMap(dict(a=1, b=2), dict(c=3))
        encoded = (
            '{"__class__":{'
            '"state":{"maps":[{"a":1,"b":2},{"c":3}]},'
            '"topic":"collections#ChainMap"}}'
        )
        self.assertTranscoding(value, encoded)

    def test_chainmap_with_tuple(self):
        value = ChainMap(dict(a=(1, 1)))
        encoded = (
            '{"__class__":{'
            '"state":{"maps":[{"a":{"__tuple__":[1,1]}}]},'
            '"topic":"collections#ChainMap"}}'
        )
        self.assertTranscoding(value, encoded)

    def test_list(self):
        value = [1, 2, 4]
        encoded = "[1,2,4]"
        self.assertTranscoding(value, encoded)

    def test_list_type(self):
        value = list
        encoded = '{"__type__":"builtins#list"}'
        self.assertTranscoding(value, encoded)

    def test_list_of_tuples(self):
        value = [(1, 1)]
        encoded = '[{"__tuple__":[1,1]}]'
        self.assertTranscoding(value, encoded)

    def test_set(self):
        value = {1, 2, 4}
        encoded = '{"__set__":[1,2,4]}'
        self.assertTranscoding(value, encoded)

    def test_set_of_tuples(self):
        value = {(1, 2)}
        encoded = '{"__set__":[{"__tuple__":[1,2]}]}'
        self.assertTranscoding(value, encoded)

    def test_set_type(self):
        value = set
        encoded = '{"__type__":"builtins#set"}'
        self.assertTranscoding(value, encoded)

    def test_frozenset(self):
        value = frozenset({1, 2, 4})
        encoded = '{"__frozenset__":[1,2,4]}'
        self.assertTranscoding(value, encoded)

    def test_frozenset_of_tuples(self):
        value = frozenset({(1, 1), (2, 2)})
        encoded = '{"__frozenset__":[{"__tuple__":[1,1]},{"__tuple__":[2,2]}]}'
        self.assertTranscoding(value, encoded)

    def test_frozenset_type(self):
        value = frozenset
        encoded = '{"__type__":"builtins#frozenset"}'
        self.assertTranscoding(value, encoded)

    def test_object(self):
        self.assertEqual(MyObjectClass(NAMESPACE_URL), MyObjectClass(NAMESPACE_URL))
        value = MyObjectClass(NAMESPACE_URL)
        encoded = (
            '{"__class__":{"state":{"a":{"UUID":"6ba7b8119dad11d18'
            '0b400c04fd430c8"}},"topic":"eventsourcing.tests.test_tr'
            'anscoding#MyObjectClass"}}'
        )
        self.assertTranscoding(value, encoded)

    def test_object_with_tuple(self):
        self.assertEqual(MyObjectClass((1, 2)), MyObjectClass((1, 2)))
        value = MyObjectClass((1, 2))
        encoded = (
            '{"__class__":{"state":{"a":{"__tuple__":[1,2]}'
            '},"topic":"eventsourcing.tests.test_tr'
            'anscoding#MyObjectClass"}}'
        )
        self.assertTranscoding(value, encoded)

    def test_object_type(self):
        value = MyObjectClass
        encoded = '{"__type__":"eventsourcing.tests.test_transcoding#MyObjectClass"}'
        self.assertTranscoding(value, encoded)

    def test_slotted_object(self):
        instance1 = MySlottedClass(a=1, b=2, c=3)
        instance2 = MySlottedClass(a=2, b=2, c=3)
        self.assertEqual(instance1, instance1)
        self.assertNotEqual(instance2, instance1)
        value = instance1
        encoded = (
            '{"__class__":{"state":{"a":1,"b":2,"c":3},"topic":'
            '"eventsourcing.tests.test_transcoding#MySlottedClass"}}'
        )
        self.assertTranscoding(value, encoded)

    def test_slotted_object_type(self):
        value = MySlottedClass
        encoded = '{"__type__":"eventsourcing.tests.test_transcoding#MySlottedClass"}'
        self.assertTranscoding(value, encoded)

    @skipIf(make_dataclass is None, "Dataclasses are not supported")
    def test_dataclass_object(self):
        value = MyDataClass(1, "2", Decimal("3.0"))
        encoded = (
            '{"__class__":{"state":{"a":1,"b":"2","c":{'
            '"__decimal__":"3.0"}},"topic":"eventsourcing.t'
            'ests.test_transcoding#MyDataClass"}}'
        )
        self.assertTranscoding(value, encoded)

    @skipIf(make_dataclass is None, "Dataclasses are not supported")
    def test_dataclass_object_type(self):
        value = MyDataClass
        encoded = '{"__type__":"eventsourcing.tests.test_transcoding#MyDataClass"}'
        self.assertTranscoding(value, encoded)

    def test_deque(self):
        value = deque([1, 3, 2])
        encoded = '{"__deque__":[1,3,2]}'
        self.assertTranscoding(value, encoded)

        value = MyDeque([1, 3, 2])
        encoded = (
            '{"__deque__":{'
            '"state":[1,3,2],'
            '"topic":"eventsourcing.tests.test_transcoding#MyDeque"'
            "}}"
        )
        self.assertTranscoding(value, encoded)

    def test_deque_with_tuples(self):
        value = deque([(1, 1), (3, 3), (2, 2)])
        encoded = (
            '{"__deque__":['
            '{"__tuple__":[1,1]},'
            '{"__tuple__":[3,3]},'
            '{"__tuple__":[2,2]}]}'
        )
        self.assertTranscoding(value, encoded)

        value = MyDeque([(1, 1), (3, 3), (2, 2)])
        encoded = (
            '{"__deque__":{"state":['
            '{"__tuple__":[1,1]},'
            '{"__tuple__":[3,3]},'
            '{"__tuple__":[2,2]}],'
            '"topic":"eventsourcing.tests.test_transcoding#MyDeque"}}'
        )
        self.assertTranscoding(value, encoded)

    def test_encode_exception_method(self):
        # Check defers to base class to raise TypeError.
        # - a type isn't supported at the moment,hence this test works
        with self.assertRaises(TypeError):
            self.encode(self.test_encode_exception_method)

    def test_encode_exception_function(self):
        # Check defers to base class to raise TypeError.
        # - a type isn't supported at the moment,hence this test works

        def my_function():
            pass

        with self.assertRaises(TypeError):
            print(self.encode(my_function))

    def test_encode_exception_lambda(self):
        # Check defers to base class to raise TypeError.
        # - a lambda function isn't supported at the moment,hence this test works

        with self.assertRaises(TypeError):
            self.encode(lambda: 1)

    def test_encode_exception_range(self):
        # Check defers to base class to raise TypeError.
        # - a type isn't supported at the moment,hence this test works

        with self.assertRaises(TypeError):
            self.encode(range(10))

    def test_decode_exception_invalid_json(self):
        # Check raises ValueError when JSON string is invalid.
        with self.assertRaises(ValueError):
            self.decode("{")

    def test_json_loads_exception(self):
        # Check raises ValueError when JSON string is invalid.
        with self.assertRaises(ValueError):
            self.decode("{")

    def assertTranscoding(self, value, encoded):
        self.assertEqual(encoded, self.encode(value))
        decoded = self.decode(encoded)
        self.assertEqual(value, decoded)
        self.assertEqual(type(value), type(decoded))

    def decode(self, encoded):
        return self.decoder.decode(encoded)

    def encode(self, value):
        return self.encoder.encode(value).decode("utf8")

    def setUp(self):
        self.encoder = ObjectJSONEncoder(sort_keys=True)
        self.decoder = ObjectJSONDecoder()


class MyDict(dict):
    pass


class MyObjectClass:
    def __init__(self, a):
        self.a = a

    def __eq__(self, other):
        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not self.__eq__(other)

    def __repr__(self):
        return "MyObjectClass(a={})".format(self.a)


MyNamedTuple = namedtuple("MyNamedTuple", field_names=["a", "b", "c"])


class MySlottedClass:
    __slots__ = ["a", "b", "c"]

    def __init__(self, a, b, c):
        self.a = a
        self.b = b
        self.c = c

    def __eq__(self, other):
        return all(getattr(self, a) == getattr(other, a) for a in self.__slots__)


if make_dataclass:
    # This decorator-style syntax with typed class attributes
    # prevents the module from loading in pypy3.5:
    #
    # @dataclass
    # class MyDataClass:
    #     a: int
    #     b: str
    #     c: Decimal
    #
    # So I needed to do this instead:
    #
    MyDataClass = make_dataclass("MyDataClass", ["a", "b", "c"])
    MyDataClass.__module__ = __name__


class Colour(Enum):
    RED = 1
    GREEN = 2
    BLUE = 3


class MyDeque(deque):
    pass



from eventsourcing.example.application import (
    close_example_application,
    get_example_application,
    init_example_application,
)
from eventsourcing.infrastructure.sqlalchemy.manager import SQLAlchemyRecordManager
from eventsourcing.infrastructure.sqlalchemy.records import IntegerSequencedNoIDRecord
from eventsourcing.tests.datastore_tests.test_sqlalchemy import (
    SQLAlchemyDatastoreTestCase,
)


class TestExampleApplicationSingleInstanceFunctions(SQLAlchemyDatastoreTestCase):
    def setUp(self):
        super(TestExampleApplicationSingleInstanceFunctions, self).setUp()
        # Setup the database.
        self.datastore.setup_connection()
        self.datastore.setup_tables()

    def tearDown(self):
        # Teardown single instance.
        close_example_application()

        # Teardown the database.
        self.datastore.drop_tables()
        self.datastore.close_connection()
        super(TestExampleApplicationSingleInstanceFunctions, self).tearDown()

    def test(self):
        self.datastore.setup_connection()
        self.datastore.setup_tables()
        record_manager = SQLAlchemyRecordManager(
            record_class=IntegerSequencedNoIDRecord, session=self.datastore.session
        )

        # Can't get the single instance before it has been constructed.
        with self.assertRaises(AssertionError):
            get_example_application()

        # Construct single instance.
        init_example_application(entity_record_manager=record_manager)

        # Can't construct single instance twice.
        with self.assertRaises(AssertionError):
            init_example_application(entity_record_manager=record_manager)

        # Get the single instance.
        app1 = get_example_application()
        app2 = get_example_application()
        self.assertEqual(id(app1), id(app2))

        # Close single instance.
        close_example_application()

        # Can't get the single instance before it has been constructed.
        with self.assertRaises(AssertionError):
            get_example_application()

        # Construct single instance.
        init_example_application(entity_record_manager=record_manager)

        # Can't construct single instance twice.
        with self.assertRaises(AssertionError):
            init_example_application(entity_record_manager=record_manager)

        # Get the single instance.
        app1 = get_example_application()
        app2 = get_example_application()
        self.assertEqual(id(app1), id(app2))



from typing import Any, Dict, Iterable, NamedTuple, Optional, Sequence
from uuid import UUID, uuid4

from axonclient.client import AxonClient, AxonEvent
from axonclient.exceptions import OutOfRangeError
from google.protobuf.internal.wire_format import INT32_MAX

from eventsourcing.infrastructure.base import RecordManagerWithNotifications


class AxonNotification(object):
    def __init__(self, tracking_token, axon_event: AxonEvent):
        self.tracking_token = tracking_token
        self.axon_event = axon_event
        self.originator_id = UUID(axon_event.aggregate_identifier)
        self.originator_version = axon_event.aggregate_sequence_number
        self.topic = axon_event.payload_type
        self.state = axon_event.payload_data

    @property
    def id(self):
        return self.tracking_token + 1


class AxonRecordManager(RecordManagerWithNotifications):
    has_integrated_snapshots = True
    can_limit_get_records = False
    can_lt_lte_get_records = False
    can_list_sequence_ids = False
    can_delete_records = False

    def __init__(self, axon_client: AxonClient, *args: Any, **kwargs: Any):
        super(AxonRecordManager, self).__init__(*args, **kwargs)
        self.axon_client = axon_client
        self.contiguous_record_ids = True
        self.notification_id_name = "id"

    def all_sequence_ids(self):
        return []

    def delete_record(self, record: Any) -> None:
        raise NotImplemented()

    def get_max_notification_id(self) -> int:
        return self.axon_client.get_last_token() + 1

    def get_notification_records(
        self,
        start: Optional[int] = None,
        stop: Optional[int] = None,
        *args: Any,
        **kwargs: Any
    ) -> Iterable:
        start = start or 0
        stop = stop or INT32_MAX
        number_of_permits = min(stop - start, INT32_MAX)
        events = self.axon_client.iter_events(
            tracking_token=start, number_of_permits=number_of_permits
        )
        for tracking_token, event in events:
            yield AxonNotification(tracking_token, axon_event=event)

    def get_record(self, sequence_id: UUID, position: int) -> Any:
        records = list(self.get_records(sequence_id, gte=position))
        if len(records):
            return records[0]
        else:
            self.raise_index_error(position)

    def get_records(
        self,
        sequence_id: UUID,
        gt: Optional[int] = None,
        gte: Optional[int] = None,
        lt: Optional[int] = None,
        lte: Optional[int] = None,
        limit: Optional[int] = None,
        query_ascending: bool = True,
        results_ascending: bool = True,
    ) -> Sequence[Any]:
        if gt is not None:
            initial_sequence_number = gt + 1
        elif gte is not None:
            initial_sequence_number = gte
        else:
            initial_sequence_number = 0

        events = self.axon_client.iter_aggregate_events(
            aggregate_id=str(sequence_id),
            initial_sequence=initial_sequence_number,
            allow_snapshots=True,
        )

        # if limit:
        #     if query_ascending:
        #         events = events[:limit]
        #     else:
        #         events = events[-limit:]

        if not results_ascending:
            events = reversed(list(events))

        return events

    def record_items(self, sequenced_items: Iterable[NamedTuple]) -> None:
        # Convert sequenced item(s) to database record(s).
        records = self.to_axon_events(sequenced_items)

        # Write records.
        self.write_records(records)

    def to_axon_events(self, items):
        for item in items:
            yield self.to_axon_event(item)

    def to_axon_event(self, item):
        message_identifier = str(uuid4())
        sequence_id = str(getattr(item, self.field_names.sequence_id))
        aggregate_sequence_number = getattr(item, self.field_names.position)
        payload_type = getattr(item, self.field_names.topic)
        payload_data = getattr(item, self.field_names.state)
        is_snapshot = payload_type.endswith("#Snapshot")
        return AxonEvent(
            message_identifier=message_identifier,
            aggregate_identifier=sequence_id,
            aggregate_sequence_number=aggregate_sequence_number,
            aggregate_type="AggregateRoot",
            timestamp=0,
            payload_type=payload_type,
            payload_revision="1",
            payload_data=payload_data,
            snapshot=is_snapshot,
            meta_data={},
        )

    def write_records(self, records):
        records = list(records)
        all_snapshots = all([r.snapshot for r in records])
        any_snapshots = any([r.snapshot for r in records])
        if any_snapshots:
            assert len(records) == 1, "Can't write more than one snapshot at once"
            if all_snapshots:
                self.axon_client.append_snapshot(records[0])
            else:
                raise AssertionError("Can't write events with snapshots")
        else:
            try:
                self.axon_client.append_event(records)
            except OutOfRangeError as e:
                self.raise_sequenced_item_conflict(e)

    def get_field_kwargs(self, item: object) -> Dict[str, Any]:
        assert isinstance(item, AxonEvent), item
        return {
            self.field_names.sequence_id: UUID(item.aggregate_identifier),
            self.field_names.position: item.aggregate_sequence_number,
            self.field_names.topic: item.payload_type,
            self.field_names.state: item.payload_data,
        }



from copy import copy
from typing import Dict
from unittest import TestCase

from eventsourcing.application.snapshotting import SnapshottingApplication
from eventsourcing.application.sqlalchemy import SQLAlchemyApplication
from eventsourcing.domain.model.aggregate import BaseAggregateRoot
from eventsourcing.domain.model.events import DomainEvent
from eventsourcing.whitehead import TEntity


class TestUpcastable(TestCase):
    """
    This test case checks that event state can be updated.
    """

    def test_upcast_recorded_state(self):
        # Upcast from original version to original version.
        UpcastableEventFixture.__class_version__ = 0
        state_v0 = {"a": 1}
        obj_state = UpcastableEventFixture.__upcast_state__(state_v0)
        self.assertEqual(obj_state["a"], 1)

        # Upcast from original version to version 1.
        UpcastableEventFixture.__class_version__ = 1
        obj_state = UpcastableEventFixture.__upcast_state__(state_v0)
        self.assertEqual(obj_state["a"], 10000)

        # Upcast from original version to version 2.
        UpcastableEventFixture.__class_version__ = 2
        obj_state = UpcastableEventFixture.__upcast_state__(state_v0)
        self.assertEqual(obj_state["a"], 1000)

        # Upcast from version 1 to version 2.
        state_v1 = {"a": 10000, "__class_version__": 1}
        obj_state = UpcastableEventFixture.__upcast_state__(state_v1)
        self.assertEqual(obj_state["a"], 1000)

        # Upcast from version 2 to version 2.
        state_v2 = {"a": 1000, "__class_version__": 2}
        obj_state = UpcastableEventFixture.__upcast_state__(state_v2)
        self.assertEqual(obj_state["a"], 1000)

    def test_event_state_has_class_version(self):
        # Construct original version.
        UpcastableEventFixture.__class_version__ = 0
        state_v0 = UpcastableEventFixture(a=1).__dict__
        # Check state doesn't have '__class_version__' attribute.
        self.assertEqual(state_v0, {"a": 1})

        # Construct version 1.
        UpcastableEventFixture.__class_version__ = 1
        state_v1 = UpcastableEventFixture(a=1).__dict__
        # Check state does have '__class_version__' attribute.
        self.assertEqual(state_v1, {"a": 1, "__class_version__": 1})

        # Construct version 2.
        UpcastableEventFixture.__class_version__ = 2
        state_v2 = UpcastableEventFixture(a=1).__dict__
        # Check state does have '__class_version__' attribute.
        self.assertEqual(state_v2, {"a": 1, "__class_version__": 2})

    def test_upcast_event_state(self):
        # Construct state with original version of the event class.
        UpcastableEventFixture.__class_version__ = 0
        state_v0 = UpcastableEventFixture(a=1).__dict__
        self.assertEqual(state_v0["a"], 1)

        # Construct state with version 1 of the event class.
        UpcastableEventFixture.__class_version__ = 1
        state_v1 = UpcastableEventFixture(a=10000).__dict__
        self.assertEqual(state_v1["a"], 10000)

        # Check original version is correctly upcast to version 1.
        state_v1_from_v0 = UpcastableEventFixture.__upcast_state__(copy(state_v0))
        self.assertEqual(state_v1, state_v1_from_v0)
        self.assertEqual(state_v0["a"], 1)

        # Check version 1 is correctly upcast to version 1.
        state_v1_from_v1 = UpcastableEventFixture.__upcast_state__(copy(state_v1))
        self.assertEqual(state_v1, state_v1_from_v1)
        self.assertEqual(state_v1["a"], 10000)

        # Construct state with version 2 of the event class.
        UpcastableEventFixture.__class_version__ = 2
        state_v2 = UpcastableEventFixture(a=1000).__dict__

        # Check original version is correctly upcast to version 2.
        state_v2_from_v0 = UpcastableEventFixture.__upcast_state__(copy(state_v0))
        self.assertEqual(state_v2, state_v2_from_v0)

        # Check version 1 is correctly upcast to version 2.
        state_v2_from_v1 = UpcastableEventFixture.__upcast_state__(copy(state_v1))
        self.assertEqual(state_v2, state_v2_from_v1)
        self.assertEqual(state_v1["a"], 10000)
        self.assertEqual(state_v0["a"], 1)

        # Check version 2 is correctly upcast to version 2.
        state_v2_from_v2 = UpcastableEventFixture.__upcast_state__(copy(state_v2))
        self.assertEqual(state_v2, state_v2_from_v2)
        self.assertEqual(state_v2["a"], 1000)
        self.assertEqual(state_v1["a"], 10000)
        self.assertEqual(state_v0["a"], 1)


class UpcastableEventFixture(DomainEvent):
    @classmethod
    def __upcast__(cls, obj_state: Dict, class_version: int):
        if class_version == 0:
            obj_state["a"] *= 10000
        elif class_version == 1:
            obj_state["a"] /= 10
        return obj_state


class TestUpcastingActiveWhenStoringAndRetrievingEvents(TestCase):
    def tearDown(self) -> None:
        # Reset class versions (in case running in suite).
        MultiVersionAggregateFixture.Triggered.__class_version__ = 0
        MultiVersionAggregateFixture.__class_version__ = 0
        del MultiVersionAggregateFixture.Triggered.__upcast__
        del MultiVersionAggregateFixture.Triggered.mutate

    def test_aggregate_state_after_versioning_events(self):
        """
        Mimick evolution of an event class, from original to version 2.

        Could/should just test reconstruct_object() calls __upcast_state__,
        but this test is decoupled from implementation, and simply checks
        upcasting is active when storing and retrieving aggregate state.
        """
        app = SQLAlchemyApplication(persist_event_type=BaseAggregateRoot.Event)
        with app:
            my_aggregate = MultiVersionAggregateFixture.__create__()
            # Trigger without any kwargs (original behaviour).
            my_aggregate.trigger()
            my_aggregate.__save__()

            # Check the default values are used for value and units.
            copy = app.repository[my_aggregate.id]
            assert isinstance(copy, MultiVersionAggregateFixture)
            self.assertEqual(copy.value, None)  # still uninitialised
            self.assertEqual(copy.units, None)  # still uninitialised

            # Increase the version of the event class (introduce 'value' attribute).
            MultiVersionAggregateFixture.Triggered.__class_version__ = 1
            MultiVersionAggregateFixture.Triggered.__upcast__ = (
                MultiVersionAggregateFixture.Triggered.upcast_v1
            )
            MultiVersionAggregateFixture.Triggered.mutate = (
                MultiVersionAggregateFixture.Triggered.mutate_v1
            )

            # Check v1 code can read original event version (backwards compatibility).
            copy = app.repository[my_aggregate.id]
            assert isinstance(copy, MultiVersionAggregateFixture)
            self.assertEqual(copy.value, 0)  # gets default
            self.assertEqual(copy.units, None)  # still uninitialised

            # Trigger with value kwarg (first version behaviour).
            my_aggregate.trigger(value=10)
            my_aggregate.__save__()

            # Check v1 code can read original and first event versions.
            copy = app.repository[my_aggregate.id]
            assert isinstance(copy, MultiVersionAggregateFixture)
            self.assertEqual(copy.value, 10)  # gets set value
            self.assertEqual(copy.units, None)  # still uninitialised

            # Increase the version of the event class.
            MultiVersionAggregateFixture.Triggered.__class_version__ = 2
            MultiVersionAggregateFixture.Triggered.__upcast__ = (
                MultiVersionAggregateFixture.Triggered.upcast_v2
            )
            MultiVersionAggregateFixture.Triggered.mutate = (
                MultiVersionAggregateFixture.Triggered.mutate_v2
            )

            # Check v2 code can read original and first event versions.
            copy = app.repository[my_aggregate.id]
            assert isinstance(copy, MultiVersionAggregateFixture)
            self.assertEqual(copy.value, 10)
            self.assertEqual(copy.units, "")

            # Trigger with value and units kwargs (second version behaviour).
            my_aggregate.trigger(value=20, units="mm")
            my_aggregate.__save__()

            # Check v2 code can read original, first and second event versions.
            copy = app.repository[my_aggregate.id]
            assert isinstance(copy, MultiVersionAggregateFixture)
            self.assertEqual(copy.value, 20)
            self.assertEqual(copy.units, "mm")

            # Check original code reads later event versions (forwards compatibility).
            MultiVersionAggregateFixture.Triggered.__class_version__ = 0
            del MultiVersionAggregateFixture.Triggered.__upcast__
            del MultiVersionAggregateFixture.Triggered.mutate
            copy = app.repository[my_aggregate.id]
            assert isinstance(copy, MultiVersionAggregateFixture)
            self.assertEqual(copy.value, None)  # still uninitialised
            self.assertEqual(copy.units, None)  # still uninitialised

            # Check v1 code reads later versions.
            MultiVersionAggregateFixture.Triggered.__class_version__ = 1
            MultiVersionAggregateFixture.Triggered.__upcast__ = (
                MultiVersionAggregateFixture.Triggered.upcast_v1
            )
            MultiVersionAggregateFixture.Triggered.mutate = (
                MultiVersionAggregateFixture.Triggered.mutate_v1
            )
            copy = app.repository[my_aggregate.id]
            assert isinstance(copy, MultiVersionAggregateFixture)
            self.assertEqual(copy.value, 20)  # observes event attribute
            self.assertEqual(copy.units, None)  # still uninitialised

    def test_snapshot_state_after_versioning_events(self):
        """
        Mimick evolution of an event class, from original to version 2.

        Could/should just test reconstruct_object() calls __upcast_state__,
        but this test is decoupled from implementation, and simply checks
        upcasting is active when storing and retrieving aggregate state.
        """
        app = SnapshottingApplication.mixin(SQLAlchemyApplication)(
            persist_event_type=BaseAggregateRoot.Event,
        )
        with app:
            my_aggregate = MultiVersionAggregateFixture.__create__()
            # Trigger without any kwargs (original behaviour).
            my_aggregate.trigger()
            my_aggregate.__save__()

            # Take a snapshot.
            app.repository.take_snapshot(my_aggregate.id)

            # Check the default values are used for value and units.
            copy = app.repository[my_aggregate.id]
            assert isinstance(copy, MultiVersionAggregateFixture)
            self.assertEqual(copy.value, None)
            self.assertEqual(copy.units, None)

            # Increase the version of the aggregate and event class.
            MultiVersionAggregateFixture.__class_version__ = 1
            MultiVersionAggregateFixture.Triggered.__class_version__ = 1
            MultiVersionAggregateFixture.Triggered.__upcast__ = (
                MultiVersionAggregateFixture.Triggered.upcast_v1
            )
            MultiVersionAggregateFixture.Triggered.mutate = (
                MultiVersionAggregateFixture.Triggered.mutate_v1
            )

            # Check v1 code upcasts snapshot (which doesnt have 'value').
            copy = app.repository[my_aggregate.id]
            assert isinstance(copy, MultiVersionAggregateFixture)
            self.assertEqual(copy.value, 0)  # gets default
            self.assertEqual(copy.units, None)

            # Trigger with value kwarg (first version behaviour).
            my_aggregate.trigger(value=10)
            my_aggregate.__save__()

            # Check v1 code can read original and first event versions.
            copy = app.repository[my_aggregate.id]
            assert isinstance(copy, MultiVersionAggregateFixture)
            self.assertEqual(copy.value, 10)
            self.assertEqual(copy.units, None)

            # Take a snapshot.
            app.repository.take_snapshot(my_aggregate.id)

            # Check the default values are used for value and units.
            copy = app.repository[my_aggregate.id]
            assert isinstance(copy, MultiVersionAggregateFixture)
            self.assertEqual(copy.value, 10)
            self.assertEqual(copy.units, None)

            # Increase the version of the aggregate and event class.
            MultiVersionAggregateFixture.__class_version__ = 2
            MultiVersionAggregateFixture.Triggered.__class_version__ = 2
            MultiVersionAggregateFixture.Triggered.__upcast__ = (
                MultiVersionAggregateFixture.Triggered.upcast_v2
            )
            MultiVersionAggregateFixture.Triggered.mutate = (
                MultiVersionAggregateFixture.Triggered.mutate_v2
            )

            # Check v2 code upcasts snapshot (which doesnt have 'units').
            copy = app.repository[my_aggregate.id]
            assert isinstance(copy, MultiVersionAggregateFixture)
            self.assertEqual(copy.value, 10)
            self.assertEqual(copy.units, "")  # gets default


class MultiVersionAggregateFixture(BaseAggregateRoot):
    DEFAULT_VALUE = 0
    DEFAULT_UNITS = ""

    @classmethod
    def __upcast__(cls, obj_state: Dict, class_version: int) -> Dict:
        if class_version == 0:
            obj_state["value"] = cls.DEFAULT_VALUE
        elif class_version == 1:
            obj_state["units"] = cls.DEFAULT_UNITS
        return obj_state

    def __init__(self, **kwargs):
        super(MultiVersionAggregateFixture, self).__init__(**kwargs)
        self.value = None
        self.units = None

    def trigger(self, **kwargs):
        self.__trigger_event__(self.Triggered, **kwargs)

    class Triggered(BaseAggregateRoot.Event):
        @property
        def value(self):
            return self.__dict__["value"]

        @property
        def units(self):
            return self.__dict__["units"]

        def mutate_v1(self, obj: TEntity) -> None:
            obj.value = self.value

        def mutate_v2(self, obj: TEntity) -> None:
            obj.value = self.value
            obj.units = self.units

        @classmethod
        def upcast_v1(cls, obj_state, class_version) -> None:
            if class_version == 0:
                obj_state["value"] = MultiVersionAggregateFixture.DEFAULT_VALUE
            return obj_state

        @classmethod
        def upcast_v2(cls, obj_state, class_version) -> None:
            if class_version == 0:
                obj_state["value"] = MultiVersionAggregateFixture.DEFAULT_VALUE
            elif class_version == 1:
                obj_state["units"] = MultiVersionAggregateFixture.DEFAULT_UNITS
            return obj_state



from eventsourcing.tests.example_application_tests import base
from eventsourcing.tests.example_application_tests.test_example_application_with_cassandra import (
    TestExampleApplicationWithCassandra,
)
from eventsourcing.tests.example_application_tests.test_example_application_with_sqlalchemy import (
    TestExampleApplicationWithSQLAlchemy,
)
from eventsourcing.utils.cipher.aes import AESCipher


class WithEncryption(base.WithExampleApplication):
    def construct_cipher(self):
        return AESCipher(cipher_key=b"0123456789abcdef")


class TestEncryptedApplicationWithCassandra(
    WithEncryption, TestExampleApplicationWithCassandra
):
    pass


class TestEncryptedApplicationWithSQLAlchemy(
    WithEncryption, TestExampleApplicationWithSQLAlchemy
):
    pass
